import struct

from ..... import emulators, platforms
from ...cstd import ArgumentType, CStdModel


class AArch64SysVModel(CStdModel):
    """Base class for C models using the AArch64 System V ABI"""

    platform = platforms.Platform(
        platforms.Architecture.AARCH64, platforms.Byteorder.LITTLE
    )
    abi = platforms.ABI.SYSTEMV

    _int_sign_mask = 0x80000000
    _int_inv_mask = 0xFFFFFFFF
    _long_sign_mask = 0x8000000000000000
    _long_inv_mask = 0xFFFFFFFFFFFFFFFF
    _long_long_sign_mask = 0x8000000000000000
    _long_long_inv_mask = 0xFFFFFFFFFFFFFFFF

    _four_byte_types = {ArgumentType.INT, ArgumentType.UINT}

    _eight_byte_types = {
        ArgumentType.LONG,
        ArgumentType.ULONG,
        ArgumentType.LONGLONG,
        ArgumentType.ULONGLONG,
        ArgumentType.SIZE_T,
        ArgumentType.SSIZE_T,
        ArgumentType.POINTER,
    }

    _four_byte_arg_regs = ["w0", "w1", "w2", "w3", "w4", "w5", "w6", "w7"]

    _eight_byte_arg_regs = [
        "x0",
        "x1",
        "x2",
        "x3",
        "x4",
        "x5",
        "x6",
        "x7",
    ]

    _soft_float = False
    _variadic_soft_float = False
    _floats_are_doubles = False
    _float_arg_regs = ["s0", "s1", "s2", "s3", "s4", "s5", "s6"]

    _double_arg_regs = ["d0", "d1", "d2", "d3", "d4", "d5"]
    _init_stack_offset = 0
    _align_stack = False
    _eight_byte_reg_size = 1
    _double_reg_size = 1
    _four_byte_stack_size = 4
    _eight_byte_stack_size = 8
    _float_stack_size = 4
    _double_stack_size = 8

    def _return_4_byte(self, emulator: emulators.Emulator, val: int) -> None:
        """Return a four-byte type"""
        emulator.write_register("w0", val)

    def _return_8_byte(self, emulator: emulators.Emulator, val: int) -> None:
        """Return an eight-byte type"""
        emulator.write_register("x0", val)

    def _return_float(self, emulator: emulators.Emulator, val: float) -> None:
        """Return a float"""
        data = struct.pack("<f", val)
        intval = int.from_bytes(data, "little")
        emulator.write_register("s0", intval)

    def _return_double(self, emulator: emulators.Emulator, val: float) -> None:
        """Return a double"""
        data = struct.pack("<d", val)
        intval = int.from_bytes(data, "little")
        emulator.write_register("d0", intval)



import random
import string
import typing

from .... import emulators, exceptions
from ....platforms import Byteorder
from ..cstd import ArgumentType, CStdModel
from ..filedesc import FDIOError, FileDescriptorManager
from .fmt_print import parse_printf_format
from .fmt_scan import FileIntake, StringIntake, handle_scanf_format
from .utils import _emu_strlen


class StdioModel(CStdModel):
    def __init__(self, address: int):
        super().__init__(address)
        self._fdmgr = FileDescriptorManager.for_platform(self.platform, self.abi)

    def _parse_mode(self, mode: str) -> typing.Tuple[bool, bool, bool, bool, bool]:
        readable = False
        writable = False
        create = False
        truncate = False
        append = True
        if mode in ("r", "rb"):
            # - Open for reading
            # - Fails if doesn't exist
            # - Cursor starts at zero
            readable = True
        elif mode in ("r+", "r+b"):
            # - Open for reading and writing
            # - Fails if doesn't exist
            # - Cursor starts at zero
            readable = True
            writable = True
        elif mode in ("w", "wb"):
            # - Open for writing
            # - Creates if doesn't exist
            # - Truncates if exists
            # - Cursor starts at zero
            writable = True
            create = True
            truncate = True
        elif mode in ("w+", "w+b"):
            # - Open for reading and writing
            # - Creates if doesn't exist
            # - Truncates if exists
            # - Cursor starts at zero
            readable = True
            writable = True
            create = True
            truncate = True
        elif mode in ("a", "ab"):
            # - Open for writing
            # - Creates if doesn't exist
            # - Cursor starts at end
            writable = True
            create = True
            append = True
        elif mode in ("a+", "a+b"):
            # - Open for reading and writing
            # - Creates if doesn't exist
            # - Start is unspecified; glibc does the beginning
            readable = True
            writable = True
            create = True
        else:
            raise FDIOError(f"Unknown mode {mode}")

        return (readable, writable, create, truncate, append)

    def _generate_tmpnam(self):
        # TODO: Doesn't actually test uniqueness
        search = string.ascii_letters + "0123456789"

        out = "/tmp/file"
        for i in range(0, 6):
            out += search[random.randint(0, len(search) - 1)]

        return out


class Fclose(StdioModel):
    name = "fclose"

    # int fclose(FILE *file);
    argument_types = [ArgumentType.POINTER]
    return_type = ArgumentType.INT

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        ptr = self.get_arg1(emulator)

        assert isinstance(ptr, int)

        try:
            fd = self._fdmgr.filestar_to_fd(ptr)
            self._fdmgr.close(fd)
            self.set_return_value(emulator, 0)
        except FDIOError:
            self.set_return_value(emulator, -1)


class Feof(StdioModel):
    name = "feof"

    # int feof(FILE *file);
    argument_types = [ArgumentType.POINTER]
    return_type = ArgumentType.INT

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        filestar = self.get_arg1(emulator)

        assert isinstance(filestar, int)

        try:
            fd = self._fdmgr.filestar_to_fd(filestar)
            file = self._fdmgr.get(fd)
        except FDIOError:
            print(f"Failed looking up filestar {filestar:x}")
            self.set_return_value(emulator, -1)
            return

        print(f"EOF: {file.eof}")
        self.set_return_value(emulator, 1 if file.eof else 0)


class Ferror(StdioModel):
    name = "ferror"

    # int ferror(FILE *file);
    argument_types = [ArgumentType.POINTER]
    return_type = ArgumentType.INT

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        # We never have errors.
        self.set_return_value(emulator, 0)


class Clearerr(StdioModel):
    name = "clearerr"

    # void clearerr(FILE *file);
    argument_types = [ArgumentType.POINTER]
    return_type = ArgumentType.VOID

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        filestar = self.get_arg1(emulator)

        assert isinstance(filestar, int)

        try:
            fd = self._fdmgr.filestar_to_fd(filestar)
            file = self._fdmgr.get(fd)
        except FDIOError:
            print(f"Failed looking up filestar {filestar:x}")
            self.set_return_value(emulator, -1)
            return

        file.eof = False


class Fflush(StdioModel):
    name = "fflush"

    # int fflush(FILE *file);
    argument_types = [ArgumentType.POINTER]
    return_type = ArgumentType.INT

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        # TODO: If you want to model something more interesting here, we can talk.
        self.set_return_value(emulator, 0)


class Fgetc(StdioModel):
    name = "fgetc"

    # int fgetc(FILE *file);
    argument_types = [ArgumentType.POINTER]
    return_type = ArgumentType.INT

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        filestar = self.get_arg1(emulator)

        assert isinstance(filestar, int)

        try:
            fd = self._fdmgr.filestar_to_fd(filestar)
            file = self._fdmgr.get(fd)
        except FDIOError:
            print(f"Failed looking up filestar {filestar:x}")
            self.set_return_value(emulator, -1)
            return

        try:
            data = file.read(1, ungetc=True)
        except Exception as e:
            print(f"Failed reading from {file.name}: {e}")
            raise e

        print(f"Returning {data[0]:x}")
        self.set_return_value(emulator, data[0])


class Fgets(StdioModel):
    name = "fgets"

    # char *fgets(char *dst, size_t size, FILE *file);
    argument_types = [ArgumentType.POINTER, ArgumentType.SIZE_T, ArgumentType.POINTER]
    return_type = ArgumentType.POINTER

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        dst = self.get_arg1(emulator)
        size = self.get_arg2(emulator)
        filestar = self.get_arg3(emulator)

        assert isinstance(dst, int)
        assert isinstance(size, int)
        assert isinstance(filestar, int)

        try:
            fd = self._fdmgr.filestar_to_fd(filestar)
            file = self._fdmgr.get(fd)
        except FDIOError:
            self.set_return_value(emulator, 0)
            return

        data = file.read_string(size)
        emulator.write_memory(dst, data)
        self.set_return_value(emulator, dst)


class Fopen(StdioModel):
    name = "fopen"

    # FILE *fopen(const char *path, const char *mode);
    argument_types = [ArgumentType.POINTER, ArgumentType.POINTER]
    return_type = ArgumentType.POINTER

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        ptr1 = self.get_arg1(emulator)
        ptr2 = self.get_arg2(emulator)

        assert isinstance(ptr1, int)
        assert isinstance(ptr2, int)

        len1 = _emu_strlen(emulator, ptr1)
        len2 = _emu_strlen(emulator, ptr2)

        bytes1 = emulator.read_memory(ptr1, len1)
        bytes2 = emulator.read_memory(ptr2, len2)

        filepath = bytes1.decode("utf-8")
        filemode = bytes2.decode("utf-8")

        # FIXME: Not all files are seekable.
        # For now, assume this one is.
        seekable = True

        try:
            readable, writable, create, truncate, append = self._parse_mode(filemode)
            fd = self._fdmgr.open(
                filepath,
                readable=readable,
                writable=writable,
                create=create,
                truncate=truncate,
                append=append,
                seekable=seekable,
            )
        except FDIOError:
            self.set_return_value(emulator, 0)
            return

        filestar = self._fdmgr.fd_to_filestar(fd)
        self.set_return_value(emulator, filestar)


class Freopen(StdioModel):
    name = "freopen"

    # FILE *freopen(const char *filename, const char *mode, FILE *stream);
    argument_types = [ArgumentType.POINTER, ArgumentType.POINTER]
    return_type = ArgumentType.POINTER

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        name = self.get_arg1(emulator)
        mode = self.get_arg2(emulator)
        filestar = self.get_arg3(emulator)

        assert isinstance(name, int)
        assert isinstance(mode, int)
        assert isinstance(filestar, int)

        len2 = _emu_strlen(emulator, mode)

        bytes2 = emulator.read_memory(mode, len2)

        filemode = bytes2.decode("utf-8")

        # FIXME: Not all files are seekable.
        # For now, assume this one is.
        seekable = True

        # Reset the access mode on the file
        try:
            fd = self._fdmgr.filestar_to_fd(filestar)
            file = self._fdmgr.get(fd)
            readable, writable, _, _, _ = self._parse_mode(filemode)
        except FDIOError:
            self.set_return_value(emulator, 0)
            return

        file.readable = readable
        file.writable = writable
        file.seekable = seekable

        # Redirect the file if name is not NULL
        if name != 0:
            len1 = _emu_strlen(emulator, name)
            bytes1 = emulator.read_memory(name, len1)
            filepath = bytes1.decode("utf-8")

            file.name = filepath

        self.set_return_value(emulator, filestar)


class Fprintf(StdioModel):
    name = "fprintf"

    # int fprintf(FILE *file, const char *fmt, ...);
    argument_types = [ArgumentType.POINTER, ArgumentType.POINTER]
    return_type = ArgumentType.INT

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        filestar = self.get_arg1(emulator)
        fmt_addr = self.get_arg2(emulator)

        assert isinstance(fmt_addr, int)

        fmt_len = _emu_strlen(emulator, fmt_addr)
        fmt_bytes = emulator.read_memory(fmt_addr, fmt_len)
        fmt = fmt_bytes.decode("utf-8")

        output = parse_printf_format(self, fmt, emulator)
        output_bytes = output.encode("utf-8")

        try:
            fd = self._fdmgr.filestar_to_fd(filestar)
            file = self._fdmgr.get(fd)
        except FDIOError:
            self.set_return_value(emulator, -1)
            return

        file.write(output_bytes)

        self.set_return_value(emulator, len(output_bytes))


class Fputc(StdioModel):
    name = "fputc"

    # int fputc(int c, FILE *file);
    argument_types = [ArgumentType.INT, ArgumentType.POINTER]
    return_type = ArgumentType.INT

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        char = self.get_arg1(emulator)
        filestar = self.get_arg2(emulator)

        assert isinstance(char, int)
        assert isinstance(filestar, int)

        try:
            fd = self._fdmgr.filestar_to_fd(filestar)
            file = self._fdmgr.get(fd)
        except FDIOError:
            self.set_return_value(emulator, -1)
            return

        file.write(bytes([char]))

        self.set_return_value(emulator, char)


class Fputs(StdioModel):
    name = "fputs"

    # int fputs(const char *str, FILE *file);
    argument_types = [ArgumentType.POINTER, ArgumentType.POINTER]
    return_type = ArgumentType.INT

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        ptr = self.get_arg1(emulator)
        filestar = self.get_arg2(emulator)

        assert isinstance(ptr, int)
        assert isinstance(filestar, int)

        strlen = _emu_strlen(emulator, ptr)
        strbytes = emulator.read_memory(ptr, strlen)

        try:
            fd = self._fdmgr.filestar_to_fd(filestar)
            file = self._fdmgr.get(fd)
        except FDIOError:
            self.set_return_value(emulator, -1)
            return

        file.write(strbytes)

        self.set_return_value(emulator, 0)


class Fread(StdioModel):
    name = "fread"

    # size_t fread(void *dst, size_t size, size_t amt, FILE *file);
    argument_types = [
        ArgumentType.POINTER,
        ArgumentType.SIZE_T,
        ArgumentType.SIZE_T,
        ArgumentType.POINTER,
    ]
    return_type = ArgumentType.SIZE_T

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        dst = self.get_arg1(emulator)
        size = self.get_arg2(emulator)
        amt = self.get_arg3(emulator)
        filestar = self.get_arg4(emulator)

        assert isinstance(dst, int)
        assert isinstance(size, int)
        assert isinstance(amt, int)
        assert isinstance(filestar, int)

        try:
            fd = self._fdmgr.filestar_to_fd(filestar)
            file = self._fdmgr.get(fd)
        except FDIOError:
            self.set_return_value(emulator, -1)
            return

        i = 0
        for i in range(0, amt):
            data = file.read(size)
            if len(data) != size:
                file.cursor -= len(data)
                break

            emulator.write_memory(dst, data)
            dst += size

        self.set_return_value(emulator, i)


class Fscanf(StdioModel):
    name = "fscanf"

    # int fscanf(FILE *, const char *fmt, ...);
    argument_types = [ArgumentType.POINTER, ArgumentType.POINTER]
    return_type = ArgumentType.INT

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        filestar = self.get_arg1(emulator)
        fmt_addr = self.get_arg2(emulator)

        assert isinstance(filestar, int)
        assert isinstance(fmt_addr, int)

        fmt_len = _emu_strlen(emulator, fmt_addr)
        fmt_bytes = emulator.read_memory(fmt_addr, fmt_len)
        fmt = fmt_bytes.decode("utf-8")

        try:
            fd = self._fdmgr.filestar_to_fd(filestar)
            file = self._fdmgr.get(fd)
        except FDIOError:
            self.set_return_value(emulator, -1)
            return

        intake = FileIntake(file)

        res = handle_scanf_format(self, intake, fmt, emulator)

        self.set_return_value(emulator, res)


class Fseek(StdioModel):
    name = "fseek"

    # int fseek(FILE *file, long int offset, int origin);
    argument_types = [ArgumentType.POINTER, ArgumentType.LONG, ArgumentType.INT]
    return_type = ArgumentType.INT

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        filestar = self.get_arg1(emulator)
        offset = self.get_arg2(emulator)
        origin = self.get_arg3(emulator)

        assert isinstance(filestar, int)
        assert isinstance(offset, int)
        assert isinstance(offset, int)

        try:
            fd = self._fdmgr.filestar_to_fd(filestar)
            file = self._fdmgr.get(fd)
        except FDIOError:
            self.set_return_value(emulator, -1)
            return

        pos = file.seek(offset, origin)
        self.set_return_value(emulator, pos)


class Ftell(StdioModel):
    name = "ftell"

    # long ftell(FILE *file);
    argument_types = [ArgumentType.POINTER]
    return_type = ArgumentType.LONG

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        filestar = self.get_arg1(emulator)

        assert isinstance(filestar, int)

        try:
            fd = self._fdmgr.filestar_to_fd(filestar)
            file = self._fdmgr.get(fd)
        except FDIOError:
            self.set_return_value(emulator, -1)

        self.set_return_value(emulator, file.cursor)


class Fgetpos(StdioModel):
    name = "fgetpos"

    # int fgetpos(FILE *file, fpos_t *pos);
    argument_types = [ArgumentType.POINTER, ArgumentType.POINTER]
    return_type = ArgumentType.INT

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        filestar = self.get_arg1(emulator)
        ptr = self.get_arg2(emulator)

        assert isinstance(filestar, int)
        assert isinstance(ptr, int)

        try:
            fd = self._fdmgr.filestar_to_fd(filestar)
            file = self._fdmgr.get(fd)
        except FDIOError:
            self.set_return_value(emulator, -1)

        # Okay, this is a pain.
        # fpos_t is platform-specific.
        # On GNU Linux, it's always a struct, and always
        # at least eight bytes.
        # One fieldis the offset.  No idea what the other is.
        #
        # If you read from inside this, I grump at you.

        if self.platform.byteorder == Byteorder.LITTLE:
            data = file.cursor.to_bytes(8, "little")
        elif self.platform.byteorder == Byteorder.BIG:
            data = file.cursor.to_bytes(8, "big")
        else:
            raise exceptions.ConfigurationError(
                f"Can't encode int for byteorder {self.platform.byteorder}"
            )

        emulator.write_memory(ptr, data)
        self.set_return_value(emulator, 0)


class Fsetpos(StdioModel):
    name = "fsetpos"

    # int ftell(FILE *file, fpos_t *pos);
    argument_types = [ArgumentType.POINTER, ArgumentType.POINTER]
    return_type = ArgumentType.INT

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        filestar = self.get_arg1(emulator)
        ptr = self.get_arg2(emulator)

        assert isinstance(filestar, int)
        assert isinstance(ptr, int)

        try:
            fd = self._fdmgr.filestar_to_fd(filestar)
            file = self._fdmgr.get(fd)
        except FDIOError:
            self.set_return_value(emulator, -1)

        # Okay, this is a pain.
        # fpos_t is platform-specific.
        # On GNU Linux, it's always a struct, and always
        # at least eight bytes.
        # One fieldis the offset.  No idea what the other is.
        #
        # If you read from inside this, I grump at you.

        data = emulator.read_memory(ptr, 8)

        if self.platform.byteorder == Byteorder.LITTLE:
            pos = int.from_bytes(data, "little")
        elif self.platform.byteorder == Byteorder.BIG:
            pos = int.from_bytes(data, "big")
        else:
            raise exceptions.ConfigurationError(
                f"Can't encode int for byteorder {self.platform.byteorder}"
            )

        file.seek(pos, 0)
        self.set_return_value(emulator, 0)


class Fwrite(StdioModel):
    name = "fwrite"

    # int fwrite(void *src, size_t size, size_t amt, FILE *file);
    argument_types = [
        ArgumentType.POINTER,
        ArgumentType.SIZE_T,
        ArgumentType.SIZE_T,
        ArgumentType.POINTER,
    ]
    return_type = ArgumentType.SIZE_T

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        src = self.get_arg1(emulator)
        size = self.get_arg2(emulator)
        amt = self.get_arg3(emulator)
        filestar = self.get_arg4(emulator)

        assert isinstance(src, int)
        assert isinstance(size, int)
        assert isinstance(amt, int)
        assert isinstance(filestar, int)

        try:
            fd = self._fdmgr.filestar_to_fd(filestar)
            file = self._fdmgr.get(fd)

            data = emulator.read_memory(src, size * amt)
            file.write(data)
        except FDIOError:
            self.set_return_value(emulator, -1)
            return

        self.set_return_value(emulator, amt)


class Getc(Fgetc):
    name = "getc"

    # NOTE: getc and fgetc behave the same.
    # getc may actually be a macro for fgetc


class Ungetc(StdioModel):
    name = "ungetc"

    # int ungetc(int c, FILE *file);
    argument_types = [ArgumentType.INT, ArgumentType.POINTER]
    return_type = ArgumentType.INT

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        char = self.get_arg1(emulator)
        filestar = self.get_arg2(emulator)

        assert isinstance(char, int)
        assert isinstance(char, int)

        try:
            fd = self._fdmgr.filestar_to_fd(filestar)
            file = self._fdmgr.get(fd)
            file.ungetc(char)
        except FDIOError:
            self.set_return_value(emulator, -1)
            return

        self.set_return_value(emulator, char)


class Getchar(StdioModel):
    name = "getchar"

    # int getchar(void);
    argument_types = []
    return_type = ArgumentType.INT

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        # Use stdin
        # TODO: If someone changes the struct pointed to by stdin, we're in trouble.
        fd = 0

        try:
            file = self._fdmgr.get(fd)
        except FDIOError:
            self.set_return_value(emulator, -1)

        data = file.read(1, ungetc=True)

        self.set_return_value(emulator, data[0])


class Gets(StdioModel):
    name = "gets"

    # char *fgets(char *dst);
    argument_types = [ArgumentType.POINTER, ArgumentType.SIZE_T, ArgumentType.POINTER]
    return_type = ArgumentType.POINTER

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        dst = self.get_arg1(emulator)

        assert isinstance(dst, int)

        try:
            fd = 0
            file = self._fdmgr.get(fd)
        except FDIOError:
            self.set_return_value(emulator, 0)
            return

        data = file.read_string()
        emulator.write_memory(dst, data)
        self.set_return_value(emulator, dst)


class Printf(StdioModel):
    name = "printf"

    # int printf(const char *fmt, ...);
    argument_types = [ArgumentType.POINTER]
    return_type = ArgumentType.INT

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        fmt_addr = self.get_arg1(emulator)

        assert isinstance(fmt_addr, int)

        fmt_len = _emu_strlen(emulator, fmt_addr)
        fmt_bytes = emulator.read_memory(fmt_addr, fmt_len)
        fmt = fmt_bytes.decode("utf-8")

        output = parse_printf_format(self, fmt, emulator)
        output_bytes = output.encode("utf-8")

        fd = 1
        try:
            file = self._fdmgr.get(fd)
        except FDIOError:
            self.set_return_value(emulator, -1)
            return

        file.write(output_bytes)

        self.set_return_value(emulator, len(output_bytes))


class Putc(Fputc):
    name = "putc"

    # NOTE: fputc and putc behave the same.
    # putc may actually be a macro for fputc.


class Putchar(StdioModel):
    name = "putchar"

    # int putchar(int c);
    argument_types = [ArgumentType.INT]
    return_type = ArgumentType.INT

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        char = self.get_arg1(emulator)

        assert isinstance(char, int)

        # Use stdout
        # TODO: If someone changes the FILE * in stdout, we're in trouble.
        fd = 1

        try:
            file = self._fdmgr.get(fd)
        except FDIOError:
            self.set_return_value(emulator, -1)
            return

        file.write(bytes([char]))

        self.set_return_value(emulator, char)


class Puts(StdioModel):
    name = "puts"

    # int puts(const char *s);
    argument_types = [ArgumentType.POINTER]
    return_type = ArgumentType.INT

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        ptr = self.get_arg1(emulator)

        assert isinstance(ptr, int)

        size = _emu_strlen(emulator, ptr)

        data = emulator.read_memory(ptr, size)

        # Use stdout
        # TODO: If someone changes the FILE * in stdout, we're in trouble.
        fd = 1

        try:
            file = self._fdmgr.get(fd)
        except FDIOError:
            self.set_return_value(emulator, -1)
            return

        file.write(data)
        file.write(b"\n")

        self.set_return_value(emulator, 0)


class Remove(StdioModel):
    name = "remove"

    # int remove(const char *path);
    argument_types = [ArgumentType.POINTER]
    return_type = ArgumentType.INT

    # No file system; just returns true
    imprecise = True

    def __init__(self, address: int):
        self._allow_imprecise = False
        super().__init__(address)

    @property
    def allow_imprecise(self) -> bool:
        return self._allow_imprecise or self._fdmgr.model_fs

    @allow_imprecise.setter
    def allow_imprecise(self, val: bool) -> None:
        self._allow_imprecise = val

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        ptr = self.get_arg1(emulator)

        assert isinstance(ptr, int)

        size = _emu_strlen(emulator, ptr)

        data = emulator.read_memory(ptr, size)

        name = data.decode("utf-8")

        if self._fdmgr.remove(name):
            self.set_return_value(emulator, 0)
        else:
            self.set_return_value(emulator, -1)


class Rename(StdioModel):
    name = "rename"

    # int rename(const char *oldpath, const char *newpath);
    argument_types = [ArgumentType.POINTER, ArgumentType.POINTER]
    return_type = ArgumentType.INT

    # No file system; just returns true
    imprecise = True

    def __init__(self, address: int):
        self._allow_imprecise = False
        super().__init__(address)

    @property
    def allow_imprecise(self) -> bool:
        return self._allow_imprecise or self._fdmgr.model_fs

    @allow_imprecise.setter
    def allow_imprecise(self, val: bool) -> None:
        self._allow_imprecise = val

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        oldptr = self.get_arg1(emulator)
        newptr = self.get_arg2(emulator)

        assert isinstance(oldptr, int)
        assert isinstance(newptr, int)

        oldsize = _emu_strlen(emulator, oldptr)
        newsize = _emu_strlen(emulator, newptr)

        olddata = emulator.read_memory(oldptr, oldsize)
        newdata = emulator.read_memory(newptr, newsize)

        old = olddata.decode("utf-8")
        new = newdata.decode("utf-8")

        try:
            self._fdmgr.rename(old, new)
            self.set_return_value(emulator, 0)
        except FDIOError:
            self.set_return_value(emulator, -1)


class Rewind(StdioModel):
    name = "rewind"

    # void rewind(FILE *file);
    argument_types = [ArgumentType.POINTER]
    return_type = ArgumentType.VOID

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        filestar = self.get_arg1(emulator)

        assert isinstance(filestar, int)

        try:
            fd = self._fdmgr.filestar_to_fd(filestar)
            file = self._fdmgr.get(fd)
        except FDIOError:
            return

        file.cursor = 0


class Scanf(StdioModel):
    name = "scanf"

    # int scanf(const char *fmt, ...);
    argument_types = [ArgumentType.POINTER]
    return_type = ArgumentType.INT

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        fmt_addr = self.get_arg1(emulator)

        assert isinstance(fmt_addr, int)

        fmt_len = _emu_strlen(emulator, fmt_addr)
        fmt_bytes = emulator.read_memory(fmt_addr, fmt_len)
        fmt = fmt_bytes.decode("utf-8")

        try:
            fd = 0
            file = self._fdmgr.get(fd)
        except FDIOError:
            self.set_return_value(emulator, -1)
            return

        intake = FileIntake(file)

        res = handle_scanf_format(self, intake, fmt, emulator)

        self.set_return_value(emulator, res)


class Snprintf(StdioModel):
    name = "snprintf"

    # int snprintf(char *dst, size_t size, const char *fmt, ...);
    argument_types = [ArgumentType.POINTER, ArgumentType.SIZE_T, ArgumentType.POINTER]
    return_type = ArgumentType.INT

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        buf_addr = self.get_arg1(emulator)
        size = self.get_arg2(emulator)
        fmt_addr = self.get_arg3(emulator)

        assert isinstance(buf_addr, int)
        assert isinstance(size, int)
        assert isinstance(fmt_addr, int)

        fmt_len = _emu_strlen(emulator, fmt_addr)
        fmt_bytes = emulator.read_memory(fmt_addr, fmt_len)
        fmt = fmt_bytes.decode("utf-8")

        output = parse_printf_format(self, fmt, emulator)
        output_bytes = output.encode("utf-8")
        output_bytes += b"\0"

        trunc_bytes = output_bytes
        if len(trunc_bytes) > size:
            trunc_bytes = trunc_bytes[: size - 1]
            trunc_bytes += b"\0"

        if buf_addr != 0:
            emulator.write_memory(buf_addr, output_bytes)

        self.set_return_value(emulator, len(output_bytes) - 1)


class Sprintf(StdioModel):
    name = "sprintf"

    # int sprintf(char *dst, const char *fmt, ...);
    argument_types = [ArgumentType.POINTER, ArgumentType.POINTER]
    return_type = ArgumentType.INT

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        buf_addr = self.get_arg1(emulator)
        fmt_addr = self.get_arg2(emulator)

        assert isinstance(buf_addr, int)
        assert isinstance(fmt_addr, int)

        fmt_len = _emu_strlen(emulator, fmt_addr)
        fmt_bytes = emulator.read_memory(fmt_addr, fmt_len)
        fmt = fmt_bytes.decode("utf-8")

        output = parse_printf_format(self, fmt, emulator)
        output_bytes = output.encode("utf-8")
        output_bytes += b"\0"

        if buf_addr != 0:
            emulator.write_memory(buf_addr, output_bytes)

        self.set_return_value(emulator, len(output_bytes) - 1)


class Sscanf(StdioModel):
    name = "sscanf"

    # int sscanf(const char *src, const char *fmt, ...);
    argument_types = [ArgumentType.POINTER, ArgumentType.POINTER]
    return_type = ArgumentType.INT

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        src_addr = self.get_arg1(emulator)
        fmt_addr = self.get_arg2(emulator)

        assert isinstance(src_addr, int)
        assert isinstance(fmt_addr, int)

        fmt_len = _emu_strlen(emulator, fmt_addr)
        fmt_bytes = emulator.read_memory(fmt_addr, fmt_len)
        fmt = fmt_bytes.decode("utf-8")

        intake = StringIntake(src_addr, emulator)

        res = handle_scanf_format(self, intake, fmt, emulator)

        self.set_return_value(emulator, res)


class Tmpfile(StdioModel):
    name = "tmpfile"

    # FILE *tmpfile(void);
    argument_types = []
    return_type = ArgumentType.POINTER

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        name = self._generate_tmpnam()

        try:
            fd = self._fdmgr.open(
                name,
                readable=True,
                writable=True,
                create=True,
                truncate=False,
                append=False,
                seekable=True,
            )
        except FDIOError:
            self.set_return_value(emulator, 0)
            return

        filestar = self._fdmgr.fd_to_filestar(fd)
        self.set_return_value(emulator, filestar)


class Tmpnam(StdioModel):
    name = "tmpnam"

    # char *tmpnam(char *name);
    argument_types = [ArgumentType.POINTER]
    return_type = ArgumentType.POINTER

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        ptr = self.get_arg1(emulator)

        assert isinstance(ptr, int)

        if ptr == 0:
            raise NotImplementedError("Using tmpnam internal buffer not supported")

        name = self._generate_tmpnam().encode("utf-8")

        emulator.write_memory(ptr, name)

        self.set_return_value(emulator, ptr)


class Vfprintf(StdioModel):
    name = "vfprintf"

    # int vfprintf(FILE *stream, const char *fmt, va_list args);
    # TODO: Figure out how to decode a va_list
    argument_types = [ArgumentType.POINTER, ArgumentType.POINTER]
    return_type = ArgumentType.INT

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        raise NotImplementedError()


class Vfscanf(StdioModel):
    name = "vfscanf"

    # int vfscanf(FILE *stream, const char *fmt, va_list args);
    # TODO: Figure out how to decode a va_list
    argument_types = [ArgumentType.POINTER, ArgumentType.POINTER]
    return_type = ArgumentType.INT

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        raise NotImplementedError()


class Vprintf(StdioModel):
    name = "vprintf"

    # int vprintf(const char *fmt, va_list args);
    # TODO: Figure out how to decode a va_list
    argument_types = [ArgumentType.POINTER]
    return_type = ArgumentType.INT

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        raise NotImplementedError()


class Vscanf(StdioModel):
    name = "vscanf"

    # int vscanf(const char *fmt, va_list args);
    # TODO: Figure out how to decode a va_list
    argument_types = [ArgumentType.POINTER]
    return_type = ArgumentType.INT

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        raise NotImplementedError()


class Vsnprintf(StdioModel):
    name = "vsnprintf"

    # int vsprintf(char *str, size_t len, const char *fmt, va_list args);
    # TODO: Figure out how to decode a va_list
    argument_types = [ArgumentType.POINTER, ArgumentType.SIZE_T, ArgumentType.POINTER]
    return_type = ArgumentType.INT

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        raise NotImplementedError()


class Vsprintf(StdioModel):
    name = "vsprintf"

    # int vsprintf(char *str, const char *fmt, va_list args);
    # TODO: Figure out how to decode a va_list
    argument_types = [ArgumentType.POINTER, ArgumentType.POINTER]
    return_type = ArgumentType.INT

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        raise NotImplementedError()


class Vsscanf(StdioModel):
    name = "vsprintf"

    # int vsscanf(char *str, const char *fmt, va_list args);
    # TODO: Figure out how to decode a va_list
    argument_types = [ArgumentType.POINTER, ArgumentType.POINTER]
    return_type = ArgumentType.INT

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        raise NotImplementedError()


__all__ = [
    "Clearerr",
    "Fclose",
    "Feof",
    "Ferror",
    "Fflush",
    "Fgetc",
    "Fgetpos",
    "Fgets",
    "Fopen",
    "Fprintf",
    "Fputc",
    "Fputs",
    "Fread",
    "Freopen",
    "Fscanf",
    "Fseek",
    "Fsetpos",
    "Ftell",
    "Fwrite",
    "Getc",
    "Getchar",
    "Gets",
    "Printf",
    "Putc",
    "Putchar",
    "Puts",
    "Remove",
    "Rename",
    "Rewind",
    "Scanf",
    "Snprintf",
    "Sprintf",
    "Sscanf",
    "Tmpfile",
    "Tmpnam",
    "Ungetc",
    "Vfprintf",
    "Vfscanf",
    "Vprintf",
    "Vscanf",
    "Vsnprintf",
    "Vsprintf",
    "Vsscanf",
]



from .unicorn import (
    UnicornEmulationError,
    UnicornEmulationExecutionError,
    UnicornEmulationMemoryReadError,
    UnicornEmulationMemoryWriteError,
    UnicornEmulator,
)

__all__ = [
    "UnicornEmulator",
    "UnicornEmulationError",
    "UnicornEmulationMemoryReadError",
    "UnicornEmulationMemoryWriteError",
    "UnicornEmulationExecutionError",
]



from ....c99 import (
    Abort,
    Abs,
    Atexit,
    Atof,
    Atoi,
    Atol,
    Atoll,
    Calloc,
    Div,
    Exit,
    Free,
    Getenv,
    LAbs,
    LDiv,
    LLAbs,
    LLDiv,
    Malloc,
    Mblen,
    Mbstowcs,
    Mbtowc,
    QSort,
    Rand,
    Realloc,
    Srand,
    System,
    Wcstombs,
    Wctomb,
)
from ..systemv import MIPS64ELSysVModel


class MIPS64ELSysVAbs(Abs, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVLAbs(LAbs, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVLLAbs(LLAbs, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVAtof(Atof, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVAtoi(Atoi, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVAtol(Atol, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVAtoll(Atoll, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVCalloc(Calloc, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVDiv(Div, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVLDiv(LDiv, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVLLDiv(LLDiv, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVExit(Exit, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVFree(Free, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVMalloc(Malloc, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVQSort(QSort, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVRand(Rand, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVRealloc(Realloc, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVSrand(Srand, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVAbort(Abort, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVAtexit(Atexit, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVGetenv(Getenv, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVMblen(Mblen, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVMbstowcs(Mbstowcs, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVMbtowc(Mbtowc, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVSystem(System, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVWcstombs(Wcstombs, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVWctomb(Wctomb, MIPS64ELSysVModel):
    pass


__all__ = [
    "MIPS64ELSysVAbs",
    "MIPS64ELSysVLAbs",
    "MIPS64ELSysVLLAbs",
    "MIPS64ELSysVAtof",
    "MIPS64ELSysVAtoi",
    "MIPS64ELSysVAtol",
    "MIPS64ELSysVAtoll",
    "MIPS64ELSysVCalloc",
    "MIPS64ELSysVDiv",
    "MIPS64ELSysVLDiv",
    "MIPS64ELSysVLLDiv",
    "MIPS64ELSysVExit",
    "MIPS64ELSysVFree",
    "MIPS64ELSysVMalloc",
    "MIPS64ELSysVQSort",
    "MIPS64ELSysVRand",
    "MIPS64ELSysVRealloc",
    "MIPS64ELSysVSrand",
    "MIPS64ELSysVAbort",
    "MIPS64ELSysVAtexit",
    "MIPS64ELSysVGetenv",
    "MIPS64ELSysVMblen",
    "MIPS64ELSysVMbstowcs",
    "MIPS64ELSysVMbtowc",
    "MIPS64ELSysVSystem",
    "MIPS64ELSysVWcstombs",
    "MIPS64ELSysVWctomb",
]



import abc
import logging
import typing

import claripy

from ... import emulators
from .. import state

logger = logging.getLogger(__name__)


class MemoryMappedModel(state.Stateful):
    """A model of a memory-mapped IO device dealing with concrete values

    This lets you specify alternate handler routines
    for reads and writes within a range of memory addressess,
    bypassing the emulator's normal memory model.
    As the name suggests, this is useful for modeling
    MMIO devices, where accessing a particular address
    actually accesses a register of a peripheral device.

    The callback routines on_read() and on_write()
    will trigger on any access within the specified range;
    one MemoryMappedModel can handle multiple registers,
    allowing you to model an entire device, or even a set of adjacent devices
    in one object.

    Arguments:
        address: Starting address of the MMIO region
        size: Size of the MMIO region in bytes
    """

    def __init__(self, address: int, size: int):
        self.address = address
        self.size = size

    @abc.abstractmethod
    def on_read(
        self, emu: emulators.Emulator, addr: int, size: int, unused: bytes
    ) -> typing.Optional[bytes]:
        """Callback handling reads from an MMIO register

        NOTE: Results must be encoded in the emulator's byte order.

        NOTE: Attempting to access memory in this MMIO range
        inside this handler will produce undefined behavior.

        Arguments:
            emu: The emulator object mediating this operation
            addr: The start address of the read operation
            size: The number of bytes read

        Returns:
            `size` bytes containing the result of the modeled read
        """
        raise NotImplementedError("Abstract method")

    @abc.abstractmethod
    def on_write(
        self, emu: emulators.Emulator, addr: int, size: int, value: bytes
    ) -> None:
        """Callback handling writes to an MMIO register

        NOTE: `value` will be encoded in the emulator's byte order.

        NOTE: Attempting to access memory in this MMIO range
        inside this handler will produce undefined behavior.

        Arguments:
            emu: The emulator object mediating this operation
            addr: The start address of the write operation
            size: The number of bytes written
            value: The bytes written
        """
        raise NotImplementedError("Abstract method")

    def extract(self, emulator: emulators.Emulator) -> None:
        logger.debug(f"Skipping extract for {self} (mmio)")

    def apply(self, emulator: emulators.Emulator) -> None:
        logger.debug(f"Hooking MMIO {self} {self.address:x}")
        emulator.map_memory(self.address, self.size)
        if self.on_read is not None:
            if not isinstance(emulator, emulators.MemoryReadHookable):
                raise NotImplementedError("Emulator does not support read hooking")
            emulator.hook_memory_read(
                self.address, self.address + self.size, self.on_read
            )
        if self.on_write is not None:
            if not isinstance(emulator, emulators.MemoryWriteHookable):
                raise NotImplementedError("Emulator does not support write hooking")
            emulator.hook_memory_write(
                self.address, self.address + self.size, self.on_write
            )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({hex(self.address)})"


class SymbolicMemoryMappedModel(state.Stateful):
    """A model of a memory-mapped IO device dealing with symbolic values

    This lets you specify alternate handler routines
    for reads and writes within a range of memory addressess,
    bypassing the emulator's normal memory model.
    As the name suggests, this is useful for modeling
    MMIO devices, where accessing a particular address
    actually accesses a register of a peripheral device.

    The callback routines on_read() and on_write()
    will trigger on any access within the specified range;
    one MemoryMappedModel can handle multiple registers,
    allowing you to model an entire device, or even a set of adjacent devices
    in one object.

    Arguments:
        address: Starting address of the MMIO region
        size: Size of the MMIO region in bytes
    """

    def __init__(self, address: int, size: int):
        self.address = address
        self.size = size

    @abc.abstractmethod
    def on_read(
        self, emu: emulators.Emulator, addr: int, size: int, unused: claripy.ast.bv.BV
    ) -> typing.Optional[claripy.ast.bv.BV]:
        """Callback handling reads from an MMIO register

        NOTE: Results must be encoded in the emulator's byte order.

        NOTE: Attempting to access memory in this MMIO range
        inside this handler will produce undefined behavior.

        Arguments:
            emu: The emulator object mediating this operation
            addr: The start address of the read operation
            size: The number of bytes read

        Returns:
            `size` bytes containing the result of the modeled read
        """
        raise NotImplementedError("Abstract method")

    @abc.abstractmethod
    def on_write(
        self, emu: emulators.Emulator, addr: int, size: int, value: claripy.ast.bv.BV
    ) -> None:
        """Callback handling writes to an MMIO register

        NOTE: `value` will be encoded in the emulator's byte order.

        NOTE: Attempting to access memory in this MMIO range
        inside this handler will produce undefined behavior.

        Arguments:
            emu: The emulator object mediating this operation
            addr: The start address of the write operation
            size: The number of bytes written
            value: The bytes written
        """
        raise NotImplementedError("Abstract method")

    def extract(self, emulator: emulators.Emulator) -> None:
        logger.debug(f"Skipping extract for {self} (mmio)")

    def apply(self, emulator: emulators.Emulator) -> None:
        logger.debug(f"Hooking MMIO {self} {self.address:x}")
        emulator.map_memory(self.address, self.size)
        if self.on_read is not None:
            if not isinstance(emulator, emulators.MemoryReadHookable):
                raise NotImplementedError("Emulator does not support read hooking")
            emulator.hook_memory_read_symbolic(
                self.address, self.address + self.size, self.on_read
            )
        if self.on_write is not None:
            if not isinstance(emulator, emulators.MemoryWriteHookable):
                raise NotImplementedError("Emulator does not support write hooking")
            emulator.hook_memory_write_symbolic(
                self.address, self.address + self.size, self.on_write
            )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({hex(self.address)})"



from __future__ import annotations

import abc
import inspect
import io
import typing
from collections.abc import Iterable


class MetadataMixin(metaclass=abc.ABCMeta):
    @property
    @abc.abstractmethod
    def name(self) -> str:
        """The name of this analysis.

        Names should be kebab-case, all lowercase, no whitespace for proper
        formatting.
        """
        pass

    @property
    @abc.abstractmethod
    def description(self) -> str:
        """A description of this analysis.

        Descriptions should be a single sentence, lowercase, with no final
        punctuation for proper formatting.
        """

        return ""

    @property
    @abc.abstractmethod
    def version(self) -> str:
        """The version string for this analysis.

        We recommend using `Semantic Versioning`_

        .. _Semantic Versioning:
            https://semver.org/
        """

        return ""


def find_subclass(
    cls: typing.Type[typing.Any],
    check: typing.Callable[[typing.Type[typing.Any]], bool],
    *args,
    **kwargs,
):
    """Find and instantiate a subclass for dependency injection

    This pattern involves finding an implementation by sematic criteria,
    rather than explicity encoding a reference to it.

    This makes for nice modular code, since the invoker
    doesn't need to get updated every time a new module is added.

    Arguments:
        cls: The class representing the interface to search
        check: Callable for testing the desired criteria
        args: Any positional/variadic args to pass to the initializer
        kwargs: Any keyword arguments to pass to the initializer

    Returns: An instance of a subclass of cls matching the criteria from check

    Raises:
        ValueError: If no subclass of cls matches the criteria
    """
    class_stack: typing.List[typing.Type[typing.Any]] = [cls]
    while len(class_stack) > 0:
        impl: typing.Type[typing.Any] = class_stack.pop(-1)

        if not inspect.isabstract(impl) and check(impl):
            return impl(*args, **kwargs)
        # __subclasses__ is not transitive;
        # need to call this on each sublclass to do a full traversal.
        class_stack.extend(impl.__subclasses__())

    raise ValueError(f"No instance of {cls} matching criteria")


class RBNode:
    def __init__(self, key, value, nil):
        self.key = key
        self.value = value
        self.is_black = False
        self.parent = None
        self.child = [nil, nil]


class RBTree(Iterable):
    """A self-balancing binary search tree

    This implements a canonical red-black tree.
    Values can be whatever you want, as long as all items
    in the tree can use the same key function.

    You can mutate values outside the tree,
    as long as you don't change their keys.

    Arguments:
        key: Function for converting values into integers for comparison
    """

    def __init__(self, key: typing.Callable[[typing.Any], int] = lambda x: x):
        self._key = key
        self._nil: RBNode = RBNode(0, None, None)
        self._nil.is_black = True
        self._root: RBNode = self._nil

    def _rotate(self, P, branch):
        G = P.parent
        S = P.child[1 - branch]

        if S is self._nil:
            raise ValueError("Can't rotate; S is nil!")

        C = S.child[branch]

        P.child[1 - branch] = C
        if C is not self._nil:
            C.parent = P

        S.child[branch] = P
        P.parent = S

        S.parent = G

        if G is not None:
            branch = 1 if P is G.child[1] else 0
            G.child[branch] = S
        else:
            self._root = S

        return S

    def insert(self, value: typing.Any) -> None:
        """Insert a value into the tree

        Arguments:
            value: The value to insert

        Raises:
            ValueError: If there's a key collision between value and something in the tree
        """
        N = RBNode(self._key(value), value, self._nil)

        # Case 0: This is the first node ever
        if self._root is self._nil:
            self._root = N
            return

        # Insert into tree normally
        P = self._root
        while True:
            if P.key > N.key:
                branch = 0
            elif P.key < N.key:
                branch = 1
            else:
                raise ValueError(f"Key collision on {value}")
            if P.child[branch] is self._nil:
                break
            P = P.child[branch]
        N.parent = P
        P.child[branch] = N

        # Rebalance the tree iteratively
        while P is not None:
            if P.is_black:
                # Case 1: P is black; we're all fine.
                break

            G = P.parent
            if G is None:
                # Case 4: P is the root and red
                # Since N is red, make P black
                P.is_black = True
                break

            # Find our uncle
            branch = 0 if G.child[0] is P else 1
            U = G.child[1 - branch]

            if U.is_black:
                if N is P.child[1 - branch]:
                    # Case 5: P is red, U is black, and N is an inner grandchild of G.
                    # Rotate the tree so it's an outer grandchild
                    self._rotate(P, branch)
                    N = P
                    P = G.child[branch]
                # Case 6: P is red, U is black, N is an outer grandchild of G
                # Rotate the tree to fix things
                self._rotate(G, 1 - branch)
                P.is_black = True
                G.is_black = False
                break

            # Case 2: P and U are red
            # Make P and U black, and G red
            P.is_black = True
            U.is_black = True
            G.is_black = False

            # Iterate one black level (2 tree levels) higher.
            N = G
            P = N.parent

        # Case 3: N is now the root and red
        # self._verify_rb(self._root)
        return

    def _get_node(self, value):
        # Helper for fetching a node,
        # or raising an error if we can't find it.
        k = self._key(value)
        N = self._root
        while N is not self._nil:
            if k == N.key:
                if value == N.value:
                    return N
                else:
                    raise ValueError(f"Key {k} had unexpected value {N.value}")
                break
            elif k < N.key:
                N = N.child[0]
            elif N.key < k:
                N = N.child[1]
        raise ValueError(f"Value {value} is not in the tree")

    def _verify_rb(self, N, c=0):
        # Verify the following properties:
        #
        # - Doubly-Linked Tree:
        #   - N.parent is None iff self._root is N
        #   - N in N.parent.child
        #
        # - Binary Search Tree:
        #   - N.child[0].key < N.key < N.child[1].key
        #
        # - Red-Black Tree:
        #   - Red nodes cannot have red children
        #   - All paths between root and nil contain the same number of black nodes
        if N is self._nil:
            return c + 1
        if N.parent is None and N is not self._root:
            raise Exception(f"DLT violation at {hex(N.key)}: Non-root has empty parent")
        if N is self._root and N.parent is not None:
            raise Exception(f"DLT violation at {hex(N.key)}: Root has non-empty parent")
        L = N.child[0]
        R = N.child[1]
        if L is not self._nil:
            if N.key <= L.key:
                raise Exception(f"BST violation at {hex(N.key)}: Left is {hex(L.key)}")
            if L.parent is not N:
                raise Exception(
                    f"DLT violation at {hex(N.key)}: Left child {hex(L.key)} parent broken"
                )
            if not N.is_black and not L.is_black:
                raise Exception(
                    f"RBT violation at {hex(N.key)}: N and left child {hex(L.key)} are red"
                )
        if R is not self._nil:
            if N.key >= R.key:
                raise Exception(f"BST violation at {hex(N.key)}: Right is {hex(R.key)}")
            if R.parent is not N:
                raise Exception(
                    f"DLT violation at {hex(N.key)}: Left child {R.key} parent broken"
                )
            if not N.is_black and not R.is_black:
                raise Exception(
                    f"RBT violation at {hex(N.key)}: N and right child {hex(R.key)} are red"
                )
        c += 1 if N.is_black else 0
        left_c = self._verify_rb(L, c)
        right_c = self._verify_rb(R, c)
        if left_c != right_c:
            raise Exception(f"RBT violation at {hex(N.key)}: {left_c} vs {right_c}")
        return left_c

    def _remove_node(self, N):
        if N.child[0] is not self._nil and N.child[1] is not self._nil:
            # Case 1: N has 2 children
            # Find in-order successor
            S = N.child[1]
            while S.child[0] is not self._nil:
                S = S.child[0]
            # Swap values
            k = N.key
            v = N.value
            N.key = S.key
            N.value = S.value
            S.key = k
            S.value = v
            # Remove the successor
            self._remove_node(S)

        elif N.child[0] is not self._nil or N.child[1] is not self._nil:
            # Case 2: N has 1 child
            C = N.child[0] if N.child[0] is not self._nil else N.child[1]
            # Make C black
            C.is_black = True

            if N.parent is None:
                # Case 2a: N is the root; no parent
                # C is now the root.
                self._root = C
                C.parent = None
            else:
                # Case 2b: N is not the root; parent
                P = N.parent
                branch = 0 if N is P.child[0] else 1
                # Prune N from the tree
                C.parent = P
                P.child[branch] = C

        elif N.parent is None:
            # Case 3: N has no children and N is the root
            # Make the root nil; we're empty
            self._root = self._nil

        elif not N.is_black:
            # Case 4: N has no children and N is red
            # Prune N out of the tree
            P = N.parent
            branch = 0 if P.child[0] is N else 1
            P.child[branch] = self._nil

        else:
            # Case 5: N has no children and N is black
            # Can't just delete the node; need to rebalance
            P = N.parent
            branch = 0 if P.child[0] is N else 1
            P.child[branch] = self._nil

            while P is not None:
                # Find our sibling, distant nephew, and close nephew
                S = P.child[1 - branch]
                D = S.child[1 - branch]
                C = S.child[branch]

                if not S.is_black:
                    # Case 5.3: S is red; P, C, D must be black
                    self._rotate(P, branch)

                    P.is_black = False
                    S.is_black = True

                    S = C
                    D = S.child[1 - branch]
                    C = S.child[branch]
                    # S is now black; handle according to 5.4, 5.5 or 5.6

                if D is not self._nil and not D.is_black:
                    # Case 5.6: S is black, D is red
                    self._rotate(P, branch)
                    S.is_black = P.is_black
                    P.is_black = True
                    D.is_black = True
                    break

                if C is not self._nil and not C.is_black:
                    # Case 5.5: S is black, C is red
                    self._rotate(S, 1 - branch)
                    S.is_black = False
                    C.is_black = True
                    D = S
                    S = C

                    # Now S is red and D is black
                    # We match case 5.6
                    self._rotate(P, branch)
                    S.is_black = P.is_black
                    P.is_black = True
                    D.is_black = True
                    break

                if not P.is_black:
                    # Case 5.4: P is red, S, C, and D are black
                    # Correct colors and we're done
                    S.is_black = False
                    P.is_black = True
                    break

                S.is_black = False
                N = P
                P = N.parent
                branch = 0 if P is None or P.child[0] is N else 1

            # Case 5.1: N is the new root; we're done.
        # self._verify_rb(self._root, 0)
        return

    def remove(self, value: typing.Any) -> None:
        """Remove a value from the tree

        Arguments:
            value: The value to remove

        Raises:
            ValueError: If `value` is not in the tree
        """
        N = self._get_node(value)
        self._remove_node(N)

    def extend(self, iterable: Iterable) -> None:
        """Add all values from an iterable to this tree

        Arguments:
            iterable: Iterable containing the values you want to add
        """
        for x in iterable:
            self.insert(x)

    def is_empty(self) -> bool:
        """Check if the tree is empty

        Returns:
            True if the tree is empty, else false.
        """
        return self._root is self._nil

    def contains(self, value: typing.Any) -> bool:
        """Check if the tree contains a value

        Arguments:
            value: The value to locate

        Returns:
            True if `value` is in the tree, else False
        """
        try:
            self._get_node(value)
            return True
        except:
            return False

    def bisect_left(self, value: typing.Any) -> typing.Optional[typing.Any]:
        """Find the value or its in-order predecessor

        Arguments:
            value: The value to search for
        Returns:
            - `value` if `value` is in the tree
            - The value with the highest key less than that of `value`
            - `None`, if there are no values with a key less than that of `value`
        """
        k = self._key(value)
        N = self._root
        # Traverse the tree to find our match
        while N is not self._nil:
            if k == N.key:
                return N.value
            elif k < N.key:
                if N.child[0] is self._nil:
                    break
                N = N.child[0]
            elif N.key < k:
                if N.child[1] is self._nil:
                    break
                N = N.child[1]

        if k < N.key:
            # Our match is to our right.
            # We need its in-order predecessor
            # This is the first ancestor for which our subtree is greater
            P = N.parent
            while P is not None:
                branch = 0 if N is P.child[0] else 1
                N = P
                P = N.parent
                if branch == 1:
                    break
            if k < N.key:
                # We are already the left-most value
                return None

        return N.value

    def bisect_right(self, value: typing.Any) -> typing.Optional[typing.Any]:
        """Find the value or its in-order successor

        Arguments:
            value: The value to search for
        Returns:
            - `value` if `value` is in the tree
            - The value with the lowest key greater than that of `value`
            - `None`, if there are no values with a key greater than that of `value`
        """
        k = self._key(value)
        N = self._root
        # Traverse the tree to find our match
        while N is not self._nil:
            if k == N.key:
                return N.value
            elif k < N.key:
                if N.child[0] is self._nil:
                    break
                N = N.child[0]
            elif k > N.key:
                if N.child[1] is self._nil:
                    break
                N = N.child[1]

        if k > N.key:
            # Our match is to our left.
            # We need its in-order successor
            # This is the first ancestor for which our subtree is less
            P = N.parent
            while P is not None:
                branch = 0 if N is P.child[0] else 1
                N = P
                P = N.parent
                if branch == 0:
                    break
            if k > N.key:
                # We are already the right-most value
                return None
        return N.value

    def _values(self, N):
        if N is not self._nil:
            for v in self._values(N.child[0]):
                yield v
            yield N.value
            for v in self._values(N.child[1]):
                yield v

    def values(self) -> typing.Any:
        """Generate values in the tree in in-order order

        Yields:
            Each value in the tree in in-order order
        """
        for v in self._values(self._root):
            yield v

    def __iter__(self):
        return iter(self.values())


def _range_collection_key(x):
    # Used to be a lambda, but can't pickle lambdas
    return x[0]


class RangeCollection(Iterable):
    """A collection of non-overlapping ranges"""

    def __init__(self):
        # Back with an RBTree keyed off the start of the range
        self._ranges = RBTree(key=_range_collection_key)

    def is_empty(self) -> bool:
        """Check if this collection is empty

        Returns:
            True iff there are no ranges in this collection
        """
        return self._ranges.is_empty()

    def contains(self, arange: typing.Tuple[int, int]) -> bool:
        """Check if a specific range overlaps any range in this collection

        Arguments:
            arange: The range to test for overlaps

        Returns:
            True iff at least one range in the collection covers at least one value in `arange`
        """

        start, end = arange
        lo = self._ranges.bisect_left(arange)
        if lo is not None:
            # There is something before start
            lo_start, lo_end = lo
            # lo_start is going to be less than or equal to start.
            # If range overlaps, start must be in lo
            if start < lo_end:
                return True

        hi = self._ranges.bisect_right(arange)
        if hi is not None:
            # There is something after start
            hi_start, hi_end = hi
            # hi_start is going to be greater than or equal to start
            # If range overlaps, hi_start must be in arange.
            if hi_start < end:
                return True

        return False

    def contains_value(self, value: int) -> bool:
        """Check if any range in this collection contains a value

        Arguments:
            value: The value to locate

        Returns:
            True iff there is a range in the collection which contains `value`
        """
        arange = (value, value + 1)

        # Only need to test left bisect
        # Value must equal start or be to the right
        lo = self._ranges.bisect_left(arange)
        if lo is not None:
            lo_start, lo_end = lo
            if value >= lo_start and value < lo_end:
                return True

        return False

    def find_closest_range(
        self, value: int
    ) -> typing.Tuple[typing.Optional[typing.Tuple[int, int]], bool]:
        """Find the range closest to a value

        Arguments:
            value: The value to locate

        Returns:
            - The closest range, or None if the collection is empty
            - True iff `value` is in that range
        """
        out = self._ranges.bisect_left((value, value + 1))
        if out is None:
            out = self._ranges.bisect_right((value, value + 1))
        return (out, out is not None and out[0] <= value < out[1])

    def add_range(self, arange: typing.Tuple[int, int]) -> None:
        """Add a range to the collection

        Arguments:
            arange: The range to insert
        """
        start, end = arange
        if start >= end:
            raise ValueError(f"Invalid range {arange}")

        lo = self._ranges.bisect_left(arange)

        if lo is not None:
            # We are not the lowest range
            lo_start, lo_end = lo
            if lo_start == start and lo_end == end:
                # We are already in the collection.  Happy happy joy joy.
                return
            # There's at least one range below us.
            # Since the tree keys off the start of ranges,
            # we can only collide with this one element
            if start >= lo_start and start <= lo_end:
                # We collide with lo.
                # Remove it from the tree, and absorb it into ourself.
                self._ranges.remove(lo)
                if lo_start < start:
                    start = lo_start
                if lo_end > end:
                    end = lo_end

        hi = self._ranges.bisect_right(arange)
        while hi is not None:
            hi_start, hi_end = hi
            if hi_start > end:
                # We don't overlap with hi.
                # We can stop slurping things up.
                break
            if hi_start < start:
                start = hi_start
            if hi_end > end:
                end = hi_end
            self._ranges.remove(hi)
            hi = self._ranges.bisect_right(arange)

        self._ranges.insert((start, end))

    def update(self, other: RangeCollection) -> None:
        """Add all ranges from another collection to this collection

        Arguments:
            other: The collection containing ranges to add
        """
        for rng in other:
            self.add_range(rng)

    def add_value(self, value: int) -> None:
        """Add a singleton range to the collection

        Arguments:
            value: The value for which to add a singleton
        """
        self.add_range((value, value + 1))

    def remove_range(self, arange: typing.Tuple[int, int]) -> None:
        """Remove any overlaps between a specific range and this collection.

        This doesn't remove a specific range;
        rather, it removes any intersections between items
        of this collection and `arange`.

        Arguments:
            arange: The range to remove
        """
        start, end = arange
        if start >= end:
            raise ValueError(f"Invalid range {arange}")
        lo = self._ranges.bisect_left(arange)
        if lo is not None:
            # We are not the lowest range
            lo_start, lo_end = lo
            if lo_start == start and lo_end == lo:
                # We exactly match an existing range
                self._ranges.remove(arange)
                return
            if start >= lo_start and end <= lo_end:
                # We collide with lo.
                # Remove lo and add the remainder back
                self._ranges.remove(lo)
                if start > lo_start:
                    # There's a bit at the beginning we need to replace
                    self._ranges.insert((lo_start, start))
                if end < lo_end:
                    # There's a bit at the end we need to replace
                    self._ranges.insert((end, lo_end))
                    # We don't need to keep going; everything will be higher.
                    return

        hi = self._ranges.bisect_right(arange)
        while hi is not None:
            hi_start, hi_end = hi
            if hi_start >= end:
                # We don't overlap with hi; we're out of ranges
                break
            self._ranges.remove(hi)
            if end < hi_end:
                # There's a bit left over at the end
                self._ranges.insert((end, hi_end))
                # We don't need to keep going; everything else will be higher.
                return
            hi = self._ranges.bisect_right(arange)

    def get_missing_ranges(
        self, arange: typing.Tuple[int, int]
    ) -> typing.List[typing.Tuple[int, int]]:
        """Find the subset of a given range not covered by this collection"""
        out = list()
        start, end = arange

        lo = self._ranges.bisect_left((start, end))
        if lo is not None:
            # There is an item below us
            lo_start, lo_end = lo
            # lo_start will be less than or equal to start,
            # so there can't be a missing range before lo.
            # We do care if there's an overlap
            if lo_end > start:
                # arange and lo overlap.  Remove the overlap
                start = lo_end

        hi = self._ranges.bisect_right((start, end))
        while hi is not None:
            # There is an item above us
            hi_start, hi_end = hi
            if hi_start >= end:
                # The item is so far above that it can't intersect
                # Anything else will be higher, so we're done.
                break
            if hi_start > start:
                # hi doesn't cover the start of arange
                # We found a missing range
                out.append((start, hi_start))
            start = hi_end
            hi = self._ranges.bisect_right((start, end))
        if start < end:
            # There's still a bit left
            out.append((start, end))
        return out

    @property
    def ranges(self) -> typing.List[typing.Tuple[int, int]]:
        """The list of ranges in order by start"""
        return list(self._ranges.values())

    def __iter__(self):
        return iter(self._ranges.values())


class SparseIO(io.BufferedIOBase):
    """Sparse memory-backed IO stream object

    BytesIO requires a contiguous bytes object.
    If you have a large, sparse memory image,
    it will gladly OOM your analysis.

    This uses an RBTree in the same manner as
    RangeCollection to maintain a non-contiguous
    sequence of bytes.
    Any data outside those ranges is assumed to be zero.

    This is used by AngrEmulator to load programs,
    and makes it possible to load sparse memory images
    covering large swaths of the address space into CLE.
    """

    def __init__(self):
        self._ranges = RBTree(key=lambda x: x[0])
        self._pos = 0
        self.size = 0

    def seekable(self) -> bool:
        return True

    def readable(self) -> bool:
        return True

    def writable(self) -> bool:
        return True

    def seek(self, pos: int, whence: int = 0) -> int:
        if not isinstance(pos, int):
            raise TypeError(f"pos must be an int; got a {type(pos)}")

        if not isinstance(whence, int):
            raise TypeError(f"whence must be an int; got a {type(whence)}")

        if whence < 0 or whence > 2:
            raise ValueError(f"Invalid whence {whence}")

        if whence == 0:
            # Relative to start of file
            self._pos = pos
        elif whence == 1:
            # Relative to current file
            self._pos += pos
        else:
            # Relative to end of file
            self._pos = self.size + pos

        return self._pos

    def read(self, size: typing.Optional[int] = -1) -> bytes:
        if size is None or size == -1:
            size = self.size
        start = self._pos
        end = start + size
        data = bytearray()

        lo = self._ranges.bisect_left((start, end, None))
        if lo is not None:
            # We are not below the lowest segment
            lo_start, lo_end, lo_data = lo
            if lo_end > start:
                # We overlap lo
                # lo_start is guaranteed to be less than or equal to start
                a = start - lo_start
                data += lo_data[a:]
                start = lo_end

        hi = self._ranges.bisect_right((start, end, None))
        while hi is not None:
            # We are not the right-most
            hi_start, hi_end, hi_data = hi
            if hi_start >= end:
                break
            if hi_start > start:
                data += b"\x00" * (hi_start - start)
                start = hi_start
            a = min(hi_end, end) - hi_start
            data += hi_data[:a]
            start = a + hi_start
            hi = self._ranges.bisect_right((start, end, None))

        if start < end:
            data += b"\x00" * (end - start)

        self._pos = end
        return bytes(data)

    def read1(self, size: int = -1) -> bytes:
        return self.read(size=size)

    def write(self, data) -> int:
        # NOTE: `data` is a bytes-like.
        # Python doesn't add a way to annotate bytes-like types
        # until 3.12
        data = bytearray(data)
        start = self._pos
        end = start + len(data)
        o_start = start
        o_end = end

        lo = self._ranges.bisect_left((start, end))
        if lo is not None:
            # We are not the lowest segment
            lo_start, lo_end, lo_data = lo
            if lo_end > start:
                # We overlap lo
                # Because of how bisect works here,
                # lo_start must be less or equal to start
                self._ranges.remove(lo)

                a = start - lo_start
                b = end - lo_start

                lo_data[a:b] = data
                data = lo_data

                start = lo_start
                if lo_end > end:
                    end = lo_end

        hi = self._ranges.bisect_right((start, end))
        while hi is not None:
            hi_start, hi_end, hi_data = hi
            # We are not the highest segment
            if hi_start >= end:
                # We do not overlap hi
                break

            # We overlap hi
            # Because of how bisect works here,
            # hi_start must be greater or equal to start
            self._ranges.remove(hi)
            if hi_end > end:
                a = end - hi_start
                data += hi_data[a:]
                end = hi_end
            hi = self._ranges.bisect_right((start, end))

        if len(data) != end - start:
            raise Exception(
                f"Buffer contains {len(data)} bytes, but have ({hex(start)}, {hex(end)}), starting with ({hex(o_start)}, {hex(o_end)})"
            )
        if end > self.size:
            self.size = end
        self._pos = end
        self._ranges.insert((start, end, data))
        return len(data)



from ..... import platforms
from .....exceptions import ConfigurationError
from ..structs import ElfRela
from .rela import ElfRelocator

R_AARCH64_ABS64 = 257  # Direct 64-bit
R_AARCH64_GLOB_DAT = 1025  # Create GOT entry
R_AARCH64_JUMP_SLOT = 1026  # Create PLT entry
R_AARCH64_RELATIVE = 1027  # Adjust by program base


class AArch64ElfRelocator(ElfRelocator):
    arch = platforms.Architecture.AARCH64
    byteorder = platforms.Byteorder.LITTLE

    def _compute_value(self, rela: ElfRela):
        if (
            rela.type == R_AARCH64_GLOB_DAT
            or rela.type == R_AARCH64_JUMP_SLOT
            or rela.type == R_AARCH64_RELATIVE
            or rela.type == R_AARCH64_ABS64
        ):
            # Different semantics, all behave the same
            val = rela.symbol.value + rela.symbol.baseaddr + rela.addend
            return val.to_bytes(8, "little")
        else:
            raise ConfigurationError(
                f"Unknown relocation type for {rela.symbol.name}: {hex(rela.type)}"
            )



import os
import typing

import claripy

from ... import emulators, exceptions, platforms
from .. import state


class Memory(state.Stateful, dict[int, state.Value]):
    """A memory region.

    A memory maps integer offsets (implicitly from the base ``address``)
    to ``Value`` objects.
    """

    def __init__(
        self,
        address: int,
        size: int,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)

        self.address: int = address
        """The start address of this memory region."""

        self.size: int = size
        """The size address of this memory region."""

    def to_bytes(self, byteorder: platforms.Byteorder) -> bytes:
        """Convert this memory region into a byte string.

        Missing/undefined space will be filled with zeros.

        Arguments:
            byteorder: Byteorder for conversion to raw bytes.

        Returns:
            Bytes for this object with the given byteorder.
        """

        result = b"\x00" * self.size
        for offset, value in self.items():
            # data = value.get_content()
            result = (
                result[:offset]
                + value.to_bytes(byteorder=byteorder)
                + result[offset + value.get_size() :]
            )

        return result

    def get_capacity(self) -> int:
        """Gets the total number of bytes this memory region can store.
        Returns:
                    The total number of bytes this memory region can store.
        """
        return self.size

    def get_used(self) -> int:
        """Gets the number of bytes written to this memory region.

        Returns:
            The number of bytes written to this memory region.
        """
        return sum([v.get_size() for v in self.values()])

    def get_size(self) -> int:
        raise NotImplementedError("You probably want get_capacity()")

    def _is_safe(self, value: state.Value):
        if (self.get_used() + value.get_size()) > self.get_capacity():
            raise ValueError("Memory is full")

    def _write_content(
        self, emulator: emulators.Emulator, address: int, value: state.Value
    ):
        # Internal stub; makes it easy for Code to switch to write_code()
        content = value.get_content()
        if isinstance(content, claripy.ast.bv.BV):
            try:
                emulator.write_memory_content(address, content)
                return
            except exceptions.SymbolicValueError:
                pass
        emulator.write_memory_content(
            address, value.to_bytes(emulator.platform.byteorder)
        )

    def apply(self, emulator: emulators.Emulator) -> None:
        emulator.map_memory(self.address, self.get_capacity())
        for offset, value in self.items():
            if value.get_content() is not None:
                self._write_content(emulator, self.address + offset, value)
            if value.get_type() is not None:
                emulator.write_memory_type(
                    self.address + offset, value.get_size(), value.get_type()
                )
            if value.get_label() is not None:
                emulator.write_memory_label(
                    self.address + offset, value.get_size(), value.get_label()
                )

    def extract(self, emulator: emulators.Emulator) -> None:
        value: state.Value
        try:
            bytes = emulator.read_memory(self.address, self.get_capacity())
            value = state.BytesValue(bytes, f"Extracted memory from {self.address}")
        except exceptions.SymbolicValueError:
            expr = emulator.read_memory_symbolic(self.address, self.get_capacity())
            value = state.SymbolicValue(
                self.get_capacity(), expr, None, f"Extracted memory from {self.address}"
            )
        except Exception as e:
            raise exceptions.EmulationError(
                f"Failed reading {hex(self.address)}"
            ) from e
        self.clear()
        self[0] = value

    def write_bytes(self, address: int, data: bytes) -> None:
        """Overwrite part of this memory region with specific bytes.
        This will fail if any sub-region of the existing memory is symbolic.

        Arguments:
            address: The address to write to.
            data: The bytes to write.
        """

        if address + len(data) > self.address + self.size:
            raise exceptions.ConfigurationError(
                f"Out of bounds: tried to read {len(data)} bytes at {hex(address)} from memory region {hex(self.address)} - {hex(self.address + self.size)}."
            )

        next_segment_address = address
        remaining_length = len(data)

        segment: state.Value
        for segment_offset, segment in sorted(self.items()):
            segment_address = self.address + segment_offset

            # fill gaps before/between existing segments
            if next_segment_address < segment_address:
                buffer = data[
                    next_segment_address - address : segment_address - address
                ]
                self[next_segment_address - self.address] = state.BytesValue(
                    buffer,
                    None,
                )
                next_segment_address += len(buffer)
                remaining_length -= len(buffer)

            # write to all overlapping segments
            if segment_address < next_segment_address:
                start_in_segment = max(segment_address, address)
                remaining_length += next_segment_address - start_in_segment
                next_segment_address = start_in_segment

            # overwrite existing segment
            if (
                remaining_length > 0
                and segment_address <= next_segment_address
                and next_segment_address < segment_address + segment.get_size()
            ):
                contents = segment.get_content()

                # check for symbolic
                if not isinstance(contents, bytes):
                    raise exceptions.SymbolicValueError(
                        f"Tried to write {len(data)} bytes at {hex(address)}. Data at {hex(segment_address)} - {hex(segment_address + segment.get_size())} is symbolic."
                    )

                offset_in_segment = next_segment_address - segment_address
                length_in_segment = min(
                    remaining_length, segment.get_size() - offset_in_segment
                )
                contents = (
                    contents[:offset_in_segment]
                    + data[
                        next_segment_address
                        - address : next_segment_address
                        - address
                        + length_in_segment
                    ]
                    + contents[offset_in_segment + length_in_segment :]
                )
                segment.set_content(contents)
                next_segment_address += length_in_segment
                remaining_length -= length_in_segment

            # check if output complete
            if remaining_length == 0:
                break

        # fill gap past end of existing segments
        if next_segment_address != address + len(data):
            self[next_segment_address - self.address] = state.BytesValue(
                data[next_segment_address - address :], None
            )

    def read_bytes(self, address: int, size: int) -> bytes:
        """Read part of this memory region.
        This will fail if any sub-region of the memory requested is symbolic or uninitialized.

        Arguments:
            address: The address to read from.
            size: The number of bytes to read.

        Returns:
            The bytes in the requested memory region.
        """

        if address + size > self.address + self.size:
            raise exceptions.ConfigurationError(
                f"Out of bounds: tried to read {size} bytes at {hex(address)} from memory region {hex(self.address)} - {hex(self.address + self.size)}."
            )

        out: bytes = b""
        next_segment_address = address
        remaining_length = size

        segment: state.Value
        for segment_offset, segment in sorted(self.items()):
            contents = segment.get_content()
            segment_address = self.address + segment_offset

            # check for symbolic
            if not isinstance(contents, bytes):
                raise exceptions.SymbolicValueError(
                    f"Tried to read {size} bytes at {hex(address)}. Data at {hex(segment_address)} - {hex(segment_address + segment.get_size())} is symbolic."
                )

            # read into output buffer
            if (
                segment_address <= next_segment_address
                and next_segment_address < segment_address + segment.get_size()
            ):
                offset_in_segment = next_segment_address - segment_address
                length_in_segment = min(
                    remaining_length, segment.get_size() - offset_in_segment
                )
                out += contents[
                    offset_in_segment : offset_in_segment + length_in_segment
                ]
                next_segment_address += length_in_segment
                remaining_length -= length_in_segment

            # check if output complete
            if remaining_length == 0:
                break

        # next segment of requested region not found
        if next_segment_address != address + size:
            raise exceptions.ConfigurationError(
                f"Tried to read {size} bytes at {hex(address)}. Data at {hex(next_segment_address)} is uninitialized."
            )

        return out

    def write_int(
        self,
        address: int,
        value: int,
        size: typing.Literal[1, 2, 4],
        endianness: platforms.Byteorder,
    ) -> None:
        """Write an integer to memory.
        This will fail if any sub-region of the existing memory is symbolic.

        Arguments:
            address: The address to write to.
            value: The integer value to write.
            size: The size of the integer in bytes.
            endianness: The byteorder of the platform.
        """

        self.write_bytes(
            address, int.to_bytes(value, size, endianness.value, signed=False)
        )

    def read_int(
        self,
        address: int,
        size: typing.Literal[1, 2, 4],
        endianness: platforms.Byteorder,
    ) -> int:
        """Read and interpret as an integer.
        This will fail if any sub-region of the memory requested is symbolic or uninitialized.

        Arguments:
            address: The address to read from.
            size: The size of the integer in bytes.
            endianness: The byteorder of the platform.

        Returns:
            The integer read from memory.
        """

        return int.from_bytes(
            self.read_bytes(address, size),
            endianness.value,
            signed=False,
        )

    def get_ranges_initialized(self) -> list[range]:
        out: list[range] = []

        for segment_offset, segment in sorted(self.items()):
            segment_start = self.address + segment_offset
            segment_end = segment_start + segment.get_size() - 1
            if len(out) > 0 and out[-1].stop + 1 == segment_end:
                segment_start = out.pop().start
            out.append(
                range(
                    segment_start,
                    segment_end,
                )
            )

        return out

    def get_ranges_uninitialized(self) -> list[range]:
        out: list[range] = []
        next_segment_start = self.address

        for segment_offset, segment in sorted(self.items()):
            segment_start = self.address + segment_offset
            segment_end = segment_start + segment.get_size()
            if next_segment_start != segment_start:
                out.append(range(next_segment_start, segment_start - 1))
            next_segment_start = segment_end

        if next_segment_start != self.address + self.get_capacity():
            out.append(
                range(next_segment_start, self.address + self.get_capacity() - 1)
            )

        return out

    def get_ranges_symbolic(self):
        out: list[range] = []

        for segment_offset, segment in sorted(self.items()):
            if not isinstance(segment, state.SymbolicValue):
                continue
            segment_start = self.address + segment_offset
            segment_end = segment_start + segment.get_size() - 1
            if len(out) > 0 and out[-1].stop + 1 == segment_end:
                segment_start = out.pop().start
            out.append(
                range(
                    segment_start,
                    segment_end,
                )
            )

        return out

    def get_ranges_concrete(self):
        out: list[range] = []

        for segment_offset, segment in sorted(self.items()):
            if isinstance(segment, state.SymbolicValue):
                continue
            segment_start = self.address + segment_offset
            segment_end = segment_start + segment.get_size() - 1
            if len(out) > 0 and out[-1].stop + 1 == segment_end:
                segment_start = out.pop().start
            out.append(
                range(
                    segment_start,
                    segment_end,
                )
            )

        return out

    def __hash__(self):
        return super(dict, self).__hash__()


class RawMemory(Memory):
    @classmethod
    def from_bytes(cls, bytes: bytes, address: int):
        """Load from a byte array.

        Arguments:
            bytes: The bytes of the memory.
            address: The address where this memory should be loaded
               into an emulator's memory.

        Returns:
            A RawMemory contstructed from the given bytes.
        """

        memory = cls(address=address, size=len(bytes))
        memory[0] = state.BytesValue(bytes, None)

        return memory

    @classmethod
    def from_file(
        cls, file: typing.BinaryIO, address: int, label: typing.Optional[str] = None
    ):
        """Load from an open file-like object.

        Arguments:
            file: The open file-like object from which to load data.
            address: The address where this memory should be loaded
               into an emulator's memory.

        Returns:
            A RawMemory from the given file-like object.
        """

        data, size = file.read(), file.tell()

        memory = cls(address=address, size=size)
        memory[0] = state.BytesValue(data, label)

        return memory

    @classmethod
    def from_filepath(cls, path: str, address: int):
        """Load from a file.

        Arguments:
            path: The path to the file.
            address: The address where this memory should be loaded
               into an emulator's memory.

        Returns:
            A RawMemory parsed from the given file path.
        """

        path = os.path.abspath(os.path.expanduser(path))

        with open(path, "rb") as f:
            return cls.from_file(file=f, address=address)


__all__ = ["Memory", "RawMemory"]



import abc
import inspect
import typing

from .... import platforms, utils

# from ....platforms import Architecture


class UnicornMachineDef(metaclass=abc.ABCMeta):
    """Container class for Unicorn architecture-specific definitions"""

    @property
    @abc.abstractmethod
    def arch(self) -> platforms.Architecture:
        """The architecture ID"""
        raise NotImplementedError("Abstract unicorn machine def has no architecture")

    @property
    @abc.abstractmethod
    def byteorder(self) -> platforms.Byteorder:
        """The byte order"""
        raise NotImplementedError("Abstract unicorn machine def has no byteorder")

    @property
    @abc.abstractmethod
    def uc_arch(self) -> int:
        """The Unicorn architecture ID"""
        return 0

    @property
    @abc.abstractmethod
    def uc_mode(self) -> int:
        """The unicorn mode ID

        This must include an byteorder flag
        """
        return 0

    _registers: typing.Dict[str, int] = {}

    def uc_reg(self, name: str) -> int:
        """Convert a register name to unicorn constant

        This must cover all names defined in the CPU state model
        for this arch/mode/byteorder, or return 0,
        which always indicates an invalid register
        """
        if name in self._registers:
            return self._registers[name]
        else:
            raise ValueError(
                f"Unknown register for {self.arch}:{self.byteorder}: {name}"
            )

    @classmethod
    def for_platform(cls, platform: platforms.Platform):
        """Find the appropriate MachineDef for your architecture

        Arguments:
            platform: platform metadata

        Returns:
            An instance of the appropriate MachineDef

        Raises:
            ValueError: If no MachineDef subclass matches your request
        """

        try:
            return utils.find_subclass(
                cls,
                lambda x: x.arch == platform.architecture
                and x.byteorder == platform.byteorder,
            )
        except:
            raise ValueError(
                f"No machine model for {platform.architecture}:{platform.byteorder}"
            )


def populate_registers(arch_info, unicorn_consts):
    def find_uc_const(reg_name):
        ew = f"_{reg_name.upper()}"
        for name, num in inspect.getmembers(unicorn_consts):
            if name.endswith(ew) and "REG" in name:
                return (name, num)
        return None

    registers = {}
    for reg_name, info in arch_info.items():
        (base_reg_name, (start, end)) = info
        (ucstr, ucnum) = find_uc_const(reg_name)
        registers[reg_name] = (ucnum, base_reg_name, start, end)

    return registers



from ..emulator import (
    Emulator,
    FunctionHookable,
    InstructionHookable,
    MemoryReadHookable,
    MemoryWriteHookable,
)


class AbstractGhidraEmulator(
    Emulator,
    InstructionHookable,
    FunctionHookable,
    MemoryReadHookable,
    MemoryWriteHookable,
):
    """Abstract type for GhidraEmulator

    Normal users of SmallWorld shouldn't need to interact with this class.

    This is for the convenience of the factory function in `__init__.py`.
    The factory needs to annotate its return type with all the
    interfaces supported by GhidraEmulator,
    but it can't use GhidraEmulator in its signature
    because it's not imported until after the first call to the factory.
    """

    pass



from ....posix.signal import (
    BsdSignal,
    Kill,
    Killpg,
    PthreadKill,
    PthreadSigmask,
    Sigaction,
    Sigaddset,
    Sigaltstack,
    Sigdelset,
    Sigemptyset,
    Sigfillset,
    Sighold,
    Sigignore,
    Siginterrupt,
    Sigismember,
    Sigpause,
    Sigpending,
    Sigprocmask,
    Sigqueue,
    Sigrelse,
    Sigset,
    Sigsuspend,
    Sigtimedwait,
    Sigwait,
    Sigwaitinfo,
)
from ..systemv import LoongArch64SysVModel


class LoongArch64SysVBsdSignal(BsdSignal, LoongArch64SysVModel):
    pass


class LoongArch64SysVKill(Kill, LoongArch64SysVModel):
    pass


class LoongArch64SysVKillpg(Killpg, LoongArch64SysVModel):
    pass


class LoongArch64SysVPthreadKill(PthreadKill, LoongArch64SysVModel):
    pass


class LoongArch64SysVPthreadSigmask(PthreadSigmask, LoongArch64SysVModel):
    pass


class LoongArch64SysVSigaction(Sigaction, LoongArch64SysVModel):
    pass


class LoongArch64SysVSigaddset(Sigaddset, LoongArch64SysVModel):
    pass


class LoongArch64SysVSigaltstack(Sigaltstack, LoongArch64SysVModel):
    pass


class LoongArch64SysVSigdelset(Sigdelset, LoongArch64SysVModel):
    pass


class LoongArch64SysVSigemptyset(Sigemptyset, LoongArch64SysVModel):
    pass


class LoongArch64SysVSigfillset(Sigfillset, LoongArch64SysVModel):
    pass


class LoongArch64SysVSighold(Sighold, LoongArch64SysVModel):
    pass


class LoongArch64SysVSigignore(Sigignore, LoongArch64SysVModel):
    pass


class LoongArch64SysVSiginterrupt(Siginterrupt, LoongArch64SysVModel):
    pass


class LoongArch64SysVSigismember(Sigismember, LoongArch64SysVModel):
    pass


class LoongArch64SysVSigpause(Sigpause, LoongArch64SysVModel):
    pass


class LoongArch64SysVSigpending(Sigpending, LoongArch64SysVModel):
    pass


class LoongArch64SysVSigprocmask(Sigprocmask, LoongArch64SysVModel):
    pass


class LoongArch64SysVSigqueue(Sigqueue, LoongArch64SysVModel):
    pass


class LoongArch64SysVSigrelse(Sigrelse, LoongArch64SysVModel):
    pass


class LoongArch64SysVSigset(Sigset, LoongArch64SysVModel):
    pass


class LoongArch64SysVSigsuspend(Sigsuspend, LoongArch64SysVModel):
    pass


class LoongArch64SysVSigtimedwait(Sigtimedwait, LoongArch64SysVModel):
    pass


class LoongArch64SysVSigwait(Sigwait, LoongArch64SysVModel):
    pass


class LoongArch64SysVSigwaitinfo(Sigwaitinfo, LoongArch64SysVModel):
    pass


__all__ = [
    "LoongArch64SysVBsdSignal",
    "LoongArch64SysVKill",
    "LoongArch64SysVKillpg",
    "LoongArch64SysVPthreadKill",
    "LoongArch64SysVPthreadSigmask",
    "LoongArch64SysVSigaction",
    "LoongArch64SysVSigaddset",
    "LoongArch64SysVSigaltstack",
    "LoongArch64SysVSigdelset",
    "LoongArch64SysVSigemptyset",
    "LoongArch64SysVSigfillset",
    "LoongArch64SysVSighold",
    "LoongArch64SysVSigignore",
    "LoongArch64SysVSiginterrupt",
    "LoongArch64SysVSigismember",
    "LoongArch64SysVSigpause",
    "LoongArch64SysVSigpending",
    "LoongArch64SysVSigprocmask",
    "LoongArch64SysVSigqueue",
    "LoongArch64SysVSigrelse",
    "LoongArch64SysVSigset",
    "LoongArch64SysVSigsuspend",
    "LoongArch64SysVSigtimedwait",
    "LoongArch64SysVSigwait",
    "LoongArch64SysVSigwaitinfo",
]



import logging
import typing

from .....exceptions import ConfigurationError
from .....platforms import Platform
from ....cpus import CPU
from ..elf import ElfExecutable
from .prstatus import PrStatus

log = logging.getLogger(__name__)


ET_CORE = 4


class ElfCoreFile(ElfExecutable):
    """
    Extended loader to handle core-dump (ET_CORE) ELF files.
    """

    def __init__(
        self,
        file: typing.BinaryIO,
        platform: typing.Optional[Platform] = None,
        ignore_platform: bool = False,
        user_base: typing.Optional[int] = None,
        page_size: int = 0x1000,
    ):
        super().__init__(
            file=file,
            platform=platform,
            ignore_platform=ignore_platform,
            user_base=user_base,
            page_size=page_size,
        )

        assert self._elf is not None
        parsed_elf = self._elf

        if parsed_elf is None or parsed_elf.header.file_type.value != ET_CORE:
            raise ConfigurationError("This file is not an ELF core dump (ET_CORE).")

        assert self.platform is not None
        self.prstatus = PrStatus.for_platform(self.platform, parsed_elf)

    def populate_cpu(self, cpu: CPU) -> None:
        self.prstatus.populate_cpu(cpu)



from ... import platforms
from .typing import AbstractGhidraEmulator


def GhidraEmulator(platform: platforms.Platform) -> AbstractGhidraEmulator:
    """Factory for creating a GhidraEmulator

    Importing any of the pyghidra packages requires
    booting up a JVM, which takes several seconds.
    It also doesn't work if Java and Ghidra are not configured.

    Rather than requiring all SmallWorld users to sit through
    this process, export a factory method that looks
    exactly like an Emulator constructor.
    Only boot pyghidra if the factory is called,
    and only import GhidraEmulator if successful.

    See GhidraEmulator in pcode.py for the actua emulator class.

    Arguments:
        platform: The platform to use when creating the emulator

    Returns:
        A GhidraEmulator object
    """

    import pyghidra

    if not pyghidra.started():
        pyghidra.start()

    from .ghidra import GhidraEmulator as Emu

    return Emu(platform)


__all__ = ["GhidraEmulator"]



from ....c99 import (
    Asctime,
    Clock,
    Ctime,
    Difftime,
    Gmtime,
    Localtime,
    Mktime,
    Strftime,
    Time,
)
from ..systemv import MIPS64ELSysVModel


class MIPS64ELSysVTime(Time, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVLocaltime(Localtime, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVGmtime(Gmtime, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVCtime(Ctime, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVAsctime(Asctime, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVStrftime(Strftime, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVDifftime(Difftime, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVMktime(Mktime, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVClock(Clock, MIPS64ELSysVModel):
    pass


__all__ = [
    "MIPS64ELSysVTime",
    "MIPS64ELSysVLocaltime",
    "MIPS64ELSysVGmtime",
    "MIPS64ELSysVCtime",
    "MIPS64ELSysVAsctime",
    "MIPS64ELSysVStrftime",
    "MIPS64ELSysVDifftime",
    "MIPS64ELSysVMktime",
    "MIPS64ELSysVClock",
]



import copy
import logging
import typing

from ... import emulators, exceptions, hinting, state
from .. import analysis

logger = logging.getLogger(__name__)


class CodeCoverage(analysis.Analysis):
    """A simple analysis that logs jumps, calls, and returns.

    Arguments:
        num_instructions: The number of instructions to execute.
    """

    def __init__(self, *args, num_instructions: int = 10, **kwargs):
        super().__init__(*args, **kwargs)
        self.num_instructions = num_instructions

    name = "code-coverage"
    description = ""
    version = "0.0.1"

    def run(self, state: state.Machine) -> None:
        machine = copy.deepcopy(state)
        cpu = machine.get_cpu()
        emulator = emulators.UnicornEmulator(cpu.platform)
        machine.apply(emulator)
        coverage: typing.Dict[int, int] = {}
        for i in range(self.num_instructions):
            pc = emulator.read_register_content("pc")
            if pc in coverage:
                coverage[pc] += 1
            else:
                coverage[pc] = 1

            try:
                emulator.step()
            except exceptions.EmulationStop:
                break
            except exceptions.EmulationError as e:
                exhint = hinting.EmulationException(
                    message="Emulation single step raised an exception",
                    pc=pc,
                    instruction_num=i,
                    exception=str(e),
                )
                self.hinter.send(exhint)
                break
        hint = hinting.CoverageHint(message="Coverage for execution", coverage=coverage)
        self.hinter.send(hint)



import angr

from .exploration import DefaultExplorationTechnique
from .memory import DefaultMemoryPlugin
from .scratch import ExpandedScratchPlugin


def configure_default_plugins(emu):
    preset = angr.SimState._presets["default"]
    preset.add_default_plugin("sym_memory", DefaultMemoryPlugin)
    preset.add_default_plugin("scratch", ExpandedScratchPlugin)


def configure_default_strategy(emu):
    emu.mgr.use_technique(DefaultExplorationTechnique())



from .bounds import BoundedExplorationMixin
from .default import DefaultExplorationTechnique
from .terminate import TerminationExplorationMixin

__all__ = [
    "BoundedExplorationMixin",
    "DefaultExplorationTechnique",
    "TerminationExplorationMixin",
]



import abc
import typing

import angr
import archinfo

from .... import exceptions, platforms, utils


class AngrMachineDef:
    """Container class for angr architecture-specific definitions"""

    @property
    @abc.abstractmethod
    def arch(self) -> platforms.Architecture:
        """The architecture ID"""
        raise NotImplementedError("This is an abstract method.")

    @property
    @abc.abstractmethod
    def byteorder(self) -> platforms.Byteorder:
        """The byte order"""
        raise NotImplementedError("This is an abstract method.")

    @property
    @abc.abstractmethod
    def angr_arch(self) -> archinfo.arch.Arch:
        """The angr architecture to use"""
        raise NotImplementedError("This is an abstract method.")

    # Is this thumb?
    # Almost always no, but angr needs to ask.
    is_thumb: bool = False

    # The angr execution engine.
    # Setting this to "none" uses the default Vex engine.
    # This only needs to be overridden if you're a pcode machine.
    angr_engine: typing.Optional[typing.Type[angr.engines.UberEnginePcode]] = None

    # Does angr support single-instruction stepping for this ISA.
    #
    # Instructions with delay slots cannot be lifted into VEX
    # without also lifting the instruction in the delay slot.
    #
    # This flag indicates that this machine uses such instructions,
    # and is not safe to step in this manner.
    supports_single_step: bool = True

    _registers: typing.Dict[str, str]

    def angr_reg(self, name: str) -> typing.Tuple[int, int]:
        """Find the offset and size of a register in the angr state's register file."""
        if name not in self._registers:
            raise KeyError(f"Unknown register for {self.arch}:{self.byteorder}: {name}")
        name = self._registers[name]

        if name not in self.angr_arch.registers:
            raise exceptions.UnsupportedRegisterError(
                f"Register {name} not recognized by angr for {self.arch}:{self.byteorder}"
            )
        return self.angr_arch.registers[name]

    def successors(self, state: angr.SimState, **kwargs) -> typing.Any:
        """Compute successor states for this architecture

        This allows a particular machine definition
        to compensate for cases where the default
        successor computation produces inaccurate results.

        For the overwhelming majority of machine models,
        the default should be sufficient.

        The biggest case to date is to handle
        user-defined operations in pcode.
        These are treated as illegal ops by angr,
        and there is currently no way to intercept their processing.

        Arguments:
            state: The angr state for which to compute successors
            kwargs: See AngrObjectFactory.successors()

        Returns:
            The successor states of `state`.
        """
        if state.project is None:
            raise exceptions.ConfigurationError("Angr state had no project.")
        return state.project.factory.successors(state, **kwargs)

    @classmethod
    def for_platform(cls, platform: platforms.Platform):
        """Find the appropriate MachineDef for your architecture

        Arguments:
            arch: The architecture ID you want
            byteorder: The byteorderness you want

        Returns:
            An instance of the appropriate MachineDef

        Raises:
            ValueError: If no MachineDef subclass matches your request
        """
        try:
            return utils.find_subclass(
                cls,
                lambda x: x.arch == platform.architecture
                and x.byteorder == platform.byteorder,
            )
        except:
            raise ValueError(f"No machine model for {platform}")


class GhidraMachineDef(AngrMachineDef):
    """Container class for pcode-dependent angr architecture-specific definitions"""

    @property
    @abc.abstractmethod
    def pcode_language(self) -> str:
        """The pcode language ID string"""
        return ""

    @property
    def angr_arch(self) -> archinfo.arch.Arch:
        return self._angr_arch

    angr_engine: typing.Optional[
        typing.Type[angr.engines.UberEnginePcode]
    ] = angr.engines.UberEnginePcode

    def __init__(self):
        self._angr_arch = archinfo.ArchPcode(self.pcode_language)



from ..... import platforms
from .....exceptions import ConfigurationError
from ..structs import ElfRela
from .rela import ElfRelocator

R_RISCV_64 = 2  #
R_RISCV_JUMP_SLOT = 5  # Create PLT entry
R_RISCV_RELATIVE = 3  # Adjust by program base


class RISCV64ElfRelocator(ElfRelocator):
    arch = platforms.Architecture.RISCV64
    byteorder = platforms.Byteorder.LITTLE

    def _compute_value(self, rela: ElfRela):
        if (
            rela.type == R_RISCV_64
            or rela.type == R_RISCV_JUMP_SLOT
            or rela.type == R_RISCV_RELATIVE
        ):
            # Different semantics, all behave the same
            val = rela.symbol.value + rela.symbol.baseaddr + rela.addend
            return val.to_bytes(8, "little")
        else:
            raise ConfigurationError(
                "Unknown relocation type for {rela.symbol.name}: {rela.type}"
            )



from ....c99.signal import Raise, Signal
from ..systemv import MIPSSysVModel


class MIPSSysVRaise(Raise, MIPSSysVModel):
    pass


class MIPSSysVSignal(Signal, MIPSSysVModel):
    pass


__all__ = [
    "MIPSSysVRaise",
    "MIPSSysVSignal",
]



from ...emulators.angr.scratch import ExpandedScratchPlugin


class GuardTrackingScratchPlugin(ExpandedScratchPlugin):
    def __init__(self, scratch=None):
        self._guard = None
        self.guards = []
        super().__init__(scratch=scratch)
        if scratch is not None:
            self.guards.extend(scratch.guards)

    @property
    def guard(self):
        return self._guard

    @guard.setter
    def guard(self, expr):
        self._guard = expr
        if self.state is not None:
            out = (self.state._ip.concrete_value, expr)
            if len(self.guards) == 0:
                self.guards.append(out)
            elif self.guards[-1][0] != out[0]:
                self.guards.append(out)
            else:
                self.guards[-1] = out



from .aarch64 import AArch64Instruction
from .arm import ARMInstruction
from .bsid import BSIDMemoryReferenceOperand
from .instructions import Instruction, Operand, RegisterOperand
from .mips import MIPSInstruction
from .x86 import AMD64Instruction, x86Instruction

__all__ = [
    "AArch64Instruction",
    "ARMInstruction",
    "AMD64Instruction",
    "BSIDMemoryReferenceOperand",
    "Instruction",
    "MIPSInstruction",
    "Operand",
    "RegisterOperand",
    "x86Instruction",
]



from ..... import platforms
from .....exceptions import ConfigurationError
from ..structs import ElfRela
from .rela import ElfRelocator

R_386_32 = 1  # Direct 32-bit
R_386_GLOB_DAT = 6  # Create GOT entry
R_386_JUMP_SLOT = 7  # Create PLT entry
R_386_RELATIVE = 8  # Adjust by program base
R_386_NUM = 44  # This and higher aren't valid


class I386ElfRelocator(ElfRelocator):
    arch = platforms.Architecture.X86_32
    byteorder = platforms.Byteorder.LITTLE

    def _compute_value(self, rela: ElfRela):
        if (
            rela.type == R_386_GLOB_DAT
            or rela.type == R_386_JUMP_SLOT
            or rela.type == R_386_RELATIVE
            or rela.type == R_386_32
        ):
            # Different semantics, all behave the same
            val = rela.symbol.value + rela.symbol.baseaddr + rela.addend
            return val.to_bytes(4, "little")
        elif rela.type >= 0 and rela.type < R_386_NUM:
            raise ConfigurationError(
                f"Valid, but unsupported relocation for {rela.symbol.name}: {rela.type}"
            )
        else:
            raise ConfigurationError(
                f"Invalid relocation type for {rela.symbol.name}: {rela.type}"
            )



class Error(Exception):
    """Common base class for all exceptions."""


class ConfigurationError(Error):
    """Raised when there is a problem with configuration."""

    pass


class EmulationError(Error):
    """Raised when emulation fails."""

    pass


class EmulationStop(EmulationError):
    """Base class for all emulation stopping exceptions."""

    pass


class EmulationBounds(EmulationStop):
    """Raised when execution goes out of bounds."""

    pass


class EmulationExitpoint(EmulationStop):
    """Raised when execution hits an exit point."""

    pass


class UnsupportedRegisterError(EmulationError):
    """Raised if you ask for a register unsupported by the emulator"""

    pass


class SymbolicValueError(EmulationError):
    """Raised if you try to collapse a symbolic value to a concrete one"""

    pass


class UnsatError(EmulationError):
    """Raised if a symbolic expression is unsatisfiable given constraints"""

    pass


class EmulationException(EmulationError):
    """Raised when the underlying emulator fails.

    This wraps known exceptions thrown by an Emulator.

    Arguments:
        exception: The original exception thrown.
    """

    def __init__(self, exception: Exception):
        self.exception = exception

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.exception})"


class AnalysisError(Error):
    """Some kind of error in analysis."""


__all__ = [
    "Error",
    "ConfigurationError",
    "EmulationError",
    "EmulationStop",
    "EmulationBounds",
    "EmulationExitpoint",
    "EmulationException",
    "SymbolicValueError",
    "UnsatError",
    "UnsupportedRegisterError",
    "AnalysisError",
]



from .aarch64 import AArch64MachineDef
from .amd64 import AMD64MachineDef
from .arm import (
    ARMv5TMachineDef,
    ARMv6MMachineDef,
    ARMv6MThumbMachineDef,
    ARMv7AMachineDef,
    ARMv7MMachineDef,
    ARMv7RMachineDef,
)
from .i386 import i386MachineDef
from .machdef import UnicornMachineDef
from .mips import MIPSBEMachineDef, MIPSELMachineDef

__all__ = [
    "AArch64MachineDef",
    "AMD64MachineDef",
    "ARMv5TMachineDef",
    "ARMv6MMachineDef",
    "ARMv6MThumbMachineDef",
    "ARMv7AMachineDef",
    "ARMv7MMachineDef",
    "ARMv7RMachineDef",
    "i386MachineDef",
    "MIPSBEMachineDef",
    "MIPSELMachineDef",
    "UnicornMachineDef",
]



import abc
import logging
import typing

import capstone

from .. import emulators, utils

logger = logging.getLogger(__name__)


class Operand(metaclass=abc.ABCMeta):
    """An operand from an instruction."""

    @abc.abstractmethod
    def key(self, emulator: emulators.Emulator):
        """Provide a unique key for this reference.

        Arguments:
            emulator: An emulator from which to fetch a value.

        Returns:
            A key value used to reference this operand.
        """

        pass

    @abc.abstractmethod
    def concretize(
        self, emulator: emulators.Emulator
    ) -> typing.Optional[typing.Union[int, bytes]]:
        """Compute a concrete value for this operand, using an emulator.

        Arguments:
            emulator: An emulator from which to fetch a value.

        Returns:
            The concrete value of this operand.
        """

        pass


class RegisterOperand(Operand):
    """An operand from an instruction that is simply a register."""

    def __init__(self, name: str):
        self.name = name

    def key(self, emulator: emulators.Emulator):
        return self.name

    def __eq__(self, other) -> bool:
        return hash(self) == hash(other)

    def __hash__(self) -> int:
        return hash(self.__repr__())

    def concretize(self, emulator: emulators.Emulator) -> int:
        return emulator.read_register(self.name)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.name})"


class MemoryReferenceOperand(Operand):
    """An operand from an instruction which reads or writes memory."""

    def __init__(self, size: int = 4):
        self.size = size

    @abc.abstractmethod
    def address(self, emulator: emulators.Emulator) -> int:
        """Compute a concrete value for this operand.

        Arguments:
            emulator: An emulator from which to fetch a value.

        Returns:
            The concrete value of this operand.
        """

    def key(self, emulator: emulators.Emulator) -> int:
        return self.address(emulator)

    def __eq__(self, other):
        return self.__repr__() == other.__repr__()

    def __hash__(self):
        return hash(self.__repr__())

    def concretize(self, emulator: emulators.Emulator) -> typing.Optional[bytes]:
        return emulator.read_memory(self.address(emulator), self.size)


class Instruction(metaclass=abc.ABCMeta):
    """An instruction storage and semantic metadata class."""

    def __init__(
        self,
        instruction: bytes,
        address: int,
        _instruction: typing.Optional[capstone.CsInsn] = None,
    ):
        self.instruction = instruction
        self.address = address

        if _instruction is None:
            md = capstone.Cs(self.cs_arch, self.cs_mode)
            md.detail = True

            _instruction = md.disasm(instruction, address).__next__()

        self._instruction = _instruction
        self.disasm = f"{self._instruction.address:x} {self._instruction.mnemonic} {self._instruction.op_str}"

    @property
    @abc.abstractmethod
    def angr_arch(self) -> str:
        """angr architecture ID"""
        return ""

    @property
    @abc.abstractmethod
    def cs_arch(self) -> int:
        """Capstone architecture ID"""
        return 0

    @property
    @abc.abstractmethod
    def cs_mode(self) -> int:
        """Capstone mode ID"""
        return 0

    @classmethod
    def from_capstone(cls, instruction: capstone.CsInsn):
        """Construct from an existing Capstone instruction.

        Arguments:
            instruction: An existing Capstone instruction.
        """
        try:
            return utils.find_subclass(
                cls,
                check=lambda x: x.cs_arch == instruction._cs.arch
                and x.cs_mode == instruction._cs.mode,
                instruction=instruction.bytes,
                address=instruction.address,
                _instruction=instruction,
            )
        except ValueError:
            raise ValueError(
                f"No instruction format for {instruction._cs.arch}:{instruction._cs.mode}"
            )

    @classmethod
    def from_angr(cls, instruction, block, arch: str):
        """Construct from an angr disassembler instruction.

        Arguments:
            instruction: An existing angr disassembler instruction
            arch: angr architecture string
        """
        # angr's instructions don't include raw bytes.
        off = instruction.address - block.addr
        raw = block.bytes[off : off + instruction.size]
        try:
            return utils.find_subclass(
                cls,
                check=lambda x: x.angr_arch == arch,
                instruction=raw,
                address=instruction.address,
            )
        except ValueError:
            raise ValueError(f"No instruction format for {arch}")

    @classmethod
    def from_bytes(cls, raw: bytes, address: int, arch: str, mode: str):
        """Construct from a byte string."""
        try:
            return utils.find_subclass(
                cls,
                check=lambda x: x.arch == arch and x.mode == mode,
                instruction=raw,
                address=address,
            )
        except ValueError:
            raise ValueError(f"No instruction format for {arch}:{mode}")

    @abc.abstractmethod
    def _memory_reference(self, operand) -> MemoryReferenceOperand:
        pass

    @property
    def reads(self) -> typing.Set[Operand]:
        """Registers and memory references read by this instruction.

        This is a list of string register names and dictionary memory reference
        specifications (i.e., in the form `base + scale * index + offset`).
        """

        read: typing.Set[Operand] = set()

        for operand in self._instruction.operands:
            if operand.type == capstone.CS_OP_MEM and (
                not hasattr(operand, "access") or operand.access & capstone.CS_AC_READ
            ):
                read.add(self._memory_reference(operand))
            elif operand.type == capstone.CS_OP_REG and (
                not hasattr(operand, "access") or operand.access & capstone.CS_AC_READ
            ):
                read.add(RegisterOperand(self._instruction.reg_name(operand.reg)))

        return read

    @property
    def writes(self) -> typing.Set[Operand]:
        """Registers and memory references written by this instruction.

        Same format as `reads`.
        """

        write: typing.Set[Operand] = set()

        for operand in self._instruction.operands:
            if operand.type == capstone.CS_OP_MEM and (
                not hasattr(operand, "access") or operand.access & capstone.CS_AC_WRITE
            ):
                write.add(self._memory_reference(operand))
            elif operand.type == capstone.CS_OP_REG and (
                not hasattr(operand, "access") or operand.access & capstone.CS_AC_WRITE
            ):
                write.add(RegisterOperand(self._instruction.reg_name(operand.reg)))

        return write

    # def to_json(self) -> dict:
    #     return {
    #         "instruction": base64.b64encode(self.instruction).decode(),
    #         "disasm": self.disasm,
    #         "address": self.address,
    #         "arch": self.arch,
    #         "mode": self.mode,
    #     }

    # @classmethod
    # def from_json(cls, dict):
    #     if "instruction" not in dict:
    #         raise ValueError(f"malformed {cls.__name__}: {dict!r}")

    #     dict["instruction"] = base64.b64decode(dict["instruction"])

    #     return cls(**dict)

    def __repr__(self) -> str:
        string = f"{self._instruction.mnemonic} {self._instruction.op_str}".strip()

        return f"{self.__class__.__name__}(0x{self.address:x}: {string})"



from ....c99 import (
    Memchr,
    Memcmp,
    Memcpy,
    Memmove,
    Memset,
    Strcat,
    Strchr,
    Strcmp,
    Strcoll,
    Strcpy,
    Strcspn,
    Strerror,
    Strlen,
    Strncat,
    Strncmp,
    Strncpy,
    Strpbrk,
    Strrchr,
    Strspn,
    Strstr,
    Strtok,
    Strxfrm,
)
from ..systemv import RiscV64SysVModel


class RiscV64SysVMemcpy(Memcpy, RiscV64SysVModel):
    pass


class RiscV64SysVMemmove(Memmove, RiscV64SysVModel):
    pass


class RiscV64SysVStrcpy(Strcpy, RiscV64SysVModel):
    pass


class RiscV64SysVStrncpy(Strncpy, RiscV64SysVModel):
    pass


class RiscV64SysVStrcat(Strcat, RiscV64SysVModel):
    pass


class RiscV64SysVStrncat(Strncat, RiscV64SysVModel):
    pass


class RiscV64SysVMemcmp(Memcmp, RiscV64SysVModel):
    pass


class RiscV64SysVStrncmp(Strncmp, RiscV64SysVModel):
    pass


class RiscV64SysVStrcmp(Strcmp, RiscV64SysVModel):
    pass


class RiscV64SysVStrcoll(Strcoll, RiscV64SysVModel):
    pass


class RiscV64SysVStrxfrm(Strxfrm, RiscV64SysVModel):
    pass


class RiscV64SysVMemchr(Memchr, RiscV64SysVModel):
    pass


class RiscV64SysVStrchr(Strchr, RiscV64SysVModel):
    pass


class RiscV64SysVStrcspn(Strcspn, RiscV64SysVModel):
    pass


class RiscV64SysVStrpbrk(Strpbrk, RiscV64SysVModel):
    pass


class RiscV64SysVStrrchr(Strrchr, RiscV64SysVModel):
    pass


class RiscV64SysVStrspn(Strspn, RiscV64SysVModel):
    pass


class RiscV64SysVStrstr(Strstr, RiscV64SysVModel):
    pass


class RiscV64SysVStrtok(Strtok, RiscV64SysVModel):
    pass


class RiscV64SysVMemset(Memset, RiscV64SysVModel):
    pass


class RiscV64SysVStrerror(Strerror, RiscV64SysVModel):
    pass


class RiscV64SysVStrlen(Strlen, RiscV64SysVModel):
    pass


__all__ = [
    "RiscV64SysVMemcpy",
    "RiscV64SysVMemmove",
    "RiscV64SysVStrcpy",
    "RiscV64SysVStrncpy",
    "RiscV64SysVStrcat",
    "RiscV64SysVStrncat",
    "RiscV64SysVMemcmp",
    "RiscV64SysVStrncmp",
    "RiscV64SysVStrcmp",
    "RiscV64SysVStrcoll",
    "RiscV64SysVStrxfrm",
    "RiscV64SysVMemchr",
    "RiscV64SysVStrchr",
    "RiscV64SysVStrcspn",
    "RiscV64SysVStrpbrk",
    "RiscV64SysVStrrchr",
    "RiscV64SysVStrspn",
    "RiscV64SysVStrstr",
    "RiscV64SysVStrtok",
    "RiscV64SysVMemset",
    "RiscV64SysVStrerror",
    "RiscV64SysVStrlen",
]



import unicorn

from ....platforms import Architecture, Byteorder
from .machdef import UnicornMachineDef


class MIPSMachineDef(UnicornMachineDef):
    """Unicorn machine definition for mips32"""

    arch = Architecture.MIPS32

    uc_arch = unicorn.UC_ARCH_MIPS
    uc_mode = unicorn.UC_MODE_MIPS32

    def __init__(self):
        self._registers = {
            # *** General-Purpose Registers ***
            # Assembler Temporary Register
            "at": unicorn.mips_const.UC_MIPS_REG_AT,
            "1": unicorn.mips_const.UC_MIPS_REG_1,
            # Return Value Registers
            "v0": unicorn.mips_const.UC_MIPS_REG_V0,
            "2": unicorn.mips_const.UC_MIPS_REG_2,
            "v1": unicorn.mips_const.UC_MIPS_REG_V1,
            "3": unicorn.mips_const.UC_MIPS_REG_3,
            # Argument Registers
            "a0": unicorn.mips_const.UC_MIPS_REG_A0,
            "4": unicorn.mips_const.UC_MIPS_REG_4,
            "a1": unicorn.mips_const.UC_MIPS_REG_A1,
            "5": unicorn.mips_const.UC_MIPS_REG_5,
            "a2": unicorn.mips_const.UC_MIPS_REG_A2,
            "6": unicorn.mips_const.UC_MIPS_REG_6,
            "a3": unicorn.mips_const.UC_MIPS_REG_A3,
            "7": unicorn.mips_const.UC_MIPS_REG_7,
            # Temporary Registers
            "t0": unicorn.mips_const.UC_MIPS_REG_T0,
            "8": unicorn.mips_const.UC_MIPS_REG_8,
            "t1": unicorn.mips_const.UC_MIPS_REG_T1,
            "9": unicorn.mips_const.UC_MIPS_REG_9,
            "t2": unicorn.mips_const.UC_MIPS_REG_T2,
            "10": unicorn.mips_const.UC_MIPS_REG_10,
            "t3": unicorn.mips_const.UC_MIPS_REG_T3,
            "11": unicorn.mips_const.UC_MIPS_REG_11,
            "t4": unicorn.mips_const.UC_MIPS_REG_T4,
            "12": unicorn.mips_const.UC_MIPS_REG_12,
            "t5": unicorn.mips_const.UC_MIPS_REG_T5,
            "13": unicorn.mips_const.UC_MIPS_REG_13,
            "t6": unicorn.mips_const.UC_MIPS_REG_T6,
            "14": unicorn.mips_const.UC_MIPS_REG_14,
            "t7": unicorn.mips_const.UC_MIPS_REG_T7,
            "15": unicorn.mips_const.UC_MIPS_REG_15,
            "t8": unicorn.mips_const.UC_MIPS_REG_T8,
            "24": unicorn.mips_const.UC_MIPS_REG_24,
            "t9": unicorn.mips_const.UC_MIPS_REG_T9,
            "25": unicorn.mips_const.UC_MIPS_REG_25,
            # Saved Registers
            "s0": unicorn.mips_const.UC_MIPS_REG_S0,
            "16": unicorn.mips_const.UC_MIPS_REG_16,
            "s1": unicorn.mips_const.UC_MIPS_REG_S1,
            "17": unicorn.mips_const.UC_MIPS_REG_17,
            "s2": unicorn.mips_const.UC_MIPS_REG_S2,
            "18": unicorn.mips_const.UC_MIPS_REG_18,
            "s3": unicorn.mips_const.UC_MIPS_REG_S3,
            "19": unicorn.mips_const.UC_MIPS_REG_19,
            "s4": unicorn.mips_const.UC_MIPS_REG_S4,
            "20": unicorn.mips_const.UC_MIPS_REG_20,
            "s5": unicorn.mips_const.UC_MIPS_REG_S5,
            "21": unicorn.mips_const.UC_MIPS_REG_21,
            "s6": unicorn.mips_const.UC_MIPS_REG_S6,
            "22": unicorn.mips_const.UC_MIPS_REG_22,
            "s7": unicorn.mips_const.UC_MIPS_REG_S7,
            "23": unicorn.mips_const.UC_MIPS_REG_23,
            # NOTE: Register 30 used to be FP, is now also s8
            "s8": unicorn.mips_const.UC_MIPS_REG_S8,
            "fp": unicorn.mips_const.UC_MIPS_REG_FP,
            "30": unicorn.mips_const.UC_MIPS_REG_30,
            # Kernel-reserved registers
            "k0": unicorn.mips_const.UC_MIPS_REG_K0,
            "26": unicorn.mips_const.UC_MIPS_REG_26,
            "k1": unicorn.mips_const.UC_MIPS_REG_K1,
            "27": unicorn.mips_const.UC_MIPS_REG_27,
            # *** Pointer Registers ***
            # Zero Register
            "zero": unicorn.mips_const.UC_MIPS_REG_ZERO,
            "0": unicorn.mips_const.UC_MIPS_REG_0,
            # Global Pointer Register
            "gp": unicorn.mips_const.UC_MIPS_REG_GP,
            "28": unicorn.mips_const.UC_MIPS_REG_28,
            # Stack Pointer Register
            "sp": unicorn.mips_const.UC_MIPS_REG_SP,
            "29": unicorn.mips_const.UC_MIPS_REG_29,
            # Return Address Register
            "ra": unicorn.mips_const.UC_MIPS_REG_RA,
            "31": unicorn.mips_const.UC_MIPS_REG_31,
            # Program Counter
            "pc": unicorn.mips_const.UC_MIPS_REG_PC,
            # *** Floating-point Registers ***
            "f0": unicorn.mips_const.UC_MIPS_REG_INVALID,
            "f1": unicorn.mips_const.UC_MIPS_REG_INVALID,
            "f2": unicorn.mips_const.UC_MIPS_REG_INVALID,
            "f3": unicorn.mips_const.UC_MIPS_REG_INVALID,
            "f4": unicorn.mips_const.UC_MIPS_REG_INVALID,
            "f5": unicorn.mips_const.UC_MIPS_REG_INVALID,
            "f6": unicorn.mips_const.UC_MIPS_REG_INVALID,
            "f7": unicorn.mips_const.UC_MIPS_REG_INVALID,
            "f8": unicorn.mips_const.UC_MIPS_REG_INVALID,
            "f9": unicorn.mips_const.UC_MIPS_REG_INVALID,
            "f10": unicorn.mips_const.UC_MIPS_REG_INVALID,
            "f11": unicorn.mips_const.UC_MIPS_REG_INVALID,
            "f12": unicorn.mips_const.UC_MIPS_REG_INVALID,
            "f13": unicorn.mips_const.UC_MIPS_REG_INVALID,
            "f14": unicorn.mips_const.UC_MIPS_REG_INVALID,
            "f15": unicorn.mips_const.UC_MIPS_REG_INVALID,
            "f16": unicorn.mips_const.UC_MIPS_REG_INVALID,
            "f17": unicorn.mips_const.UC_MIPS_REG_INVALID,
            "f18": unicorn.mips_const.UC_MIPS_REG_INVALID,
            "f19": unicorn.mips_const.UC_MIPS_REG_INVALID,
            "f20": unicorn.mips_const.UC_MIPS_REG_INVALID,
            "f21": unicorn.mips_const.UC_MIPS_REG_INVALID,
            "f22": unicorn.mips_const.UC_MIPS_REG_INVALID,
            "f23": unicorn.mips_const.UC_MIPS_REG_INVALID,
            "f24": unicorn.mips_const.UC_MIPS_REG_INVALID,
            "f25": unicorn.mips_const.UC_MIPS_REG_INVALID,
            "f26": unicorn.mips_const.UC_MIPS_REG_INVALID,
            "f27": unicorn.mips_const.UC_MIPS_REG_INVALID,
            "f28": unicorn.mips_const.UC_MIPS_REG_INVALID,
            "f29": unicorn.mips_const.UC_MIPS_REG_INVALID,
            "f30": unicorn.mips_const.UC_MIPS_REG_INVALID,
            "f31": unicorn.mips_const.UC_MIPS_REG_INVALID,
            # *** Floating Point Control Registers ***
            # NOTE: These are taken from Sleigh, and the MIPS docs.
            # Unicorn doesn't use these names, and has a different number of registers.
            "fir": unicorn.mips_const.UC_MIPS_REG_INVALID,
            "fcsr": unicorn.mips_const.UC_MIPS_REG_INVALID,
            "fexr": unicorn.mips_const.UC_MIPS_REG_INVALID,
            "fenr": unicorn.mips_const.UC_MIPS_REG_INVALID,
            "fccr": unicorn.mips_const.UC_MIPS_REG_INVALID,
        }


class MIPSELMachineDef(MIPSMachineDef):
    """Unicorn machine definition for mips32 little-endian"""

    byteorder = Byteorder.LITTLE

    def __init__(self):
        super().__init__()
        self._registers.update(
            {
                # *** Accumulator Registers ***
                # TODO: Unicorn broke support for these in 2.0.2
                "ac0": unicorn.mips_const.UC_MIPS_REG_INVALID,
                "lo0": unicorn.mips_const.UC_MIPS_REG_INVALID,
                "hi0": unicorn.mips_const.UC_MIPS_REG_INVALID,
                "ac1": unicorn.mips_const.UC_MIPS_REG_INVALID,
                "lo1": unicorn.mips_const.UC_MIPS_REG_INVALID,
                "hi1": unicorn.mips_const.UC_MIPS_REG_INVALID,
                "ac2": unicorn.mips_const.UC_MIPS_REG_INVALID,
                "lo2": unicorn.mips_const.UC_MIPS_REG_INVALID,
                "hi2": unicorn.mips_const.UC_MIPS_REG_INVALID,
                "ac3": unicorn.mips_const.UC_MIPS_REG_INVALID,
                "lo3": unicorn.mips_const.UC_MIPS_REG_INVALID,
                "hi3": unicorn.mips_const.UC_MIPS_REG_INVALID,
            }
        )


class MIPSBEMachineDef(MIPSMachineDef):
    """Unicorn machine definition for mips32 big-endian"""

    byteorder = Byteorder.BIG

    uc_mode = unicorn.UC_MODE_MIPS32 | unicorn.UC_MODE_BIG_ENDIAN

    def __init__(self):
        super().__init__()
        self._registers.update(
            {
                # *** Accumulator Registers ***
                # TODO: Unicorn broke support for these in 2.0.2
                "ac0": unicorn.mips_const.UC_MIPS_REG_INVALID,
                "hi0": unicorn.mips_const.UC_MIPS_REG_INVALID,
                "lo0": unicorn.mips_const.UC_MIPS_REG_INVALID,
                "ac1": unicorn.mips_const.UC_MIPS_REG_INVALID,
                "hi1": unicorn.mips_const.UC_MIPS_REG_INVALID,
                "lo1": unicorn.mips_const.UC_MIPS_REG_INVALID,
                "ac2": unicorn.mips_const.UC_MIPS_REG_INVALID,
                "hi2": unicorn.mips_const.UC_MIPS_REG_INVALID,
                "lo2": unicorn.mips_const.UC_MIPS_REG_INVALID,
                "ac3": unicorn.mips_const.UC_MIPS_REG_INVALID,
                "hi3": unicorn.mips_const.UC_MIPS_REG_INVALID,
                "lo3": unicorn.mips_const.UC_MIPS_REG_INVALID,
            }
        )



from ..... import platforms
from .....exceptions import ConfigurationError
from ..structs import ElfRela
from .rela import ElfRelocator

R_MIPS_NONE = 0
R_MIPS_32 = 2  # 32-bit direct
R_MIPS_REL32 = 3
R_MIPS_64 = 18  # 64-bit direct


class MIPSElfRelocator(ElfRelocator):
    arch = platforms.Architecture.MIPS32
    byteorder = platforms.Byteorder.BIG

    def _compute_value(self, rela: ElfRela):
        # Unpack a MIPS64 relocation.
        #
        # It's not possible to build a 64-bit address in one instruction in MIPS64.
        # Rather than tell the linker to live with it,
        # MIPS64 relocation entries can define up to three nested relocations,
        # with relocations 2 and 3 acting as if the symbol value were the result
        # of the previous relocation.
        #
        # There's a fourth argument, which is a parameter to relocation 2, I think.
        # Don't want to waste those bytes...
        rela_types = []
        # rela_special = (rela.type >> 24) & 0xFF
        for i in range(0, 3):
            rela_types.append((rela.type >> (i * 8)) & 0xFF)

        val = rela.symbol.value + rela.symbol.baseaddr
        is_64 = False
        for i in range(0, 3):
            rela_type = rela_types[i]
            if rela_type == R_MIPS_NONE:
                break
            elif rela_type == R_MIPS_32 or rela_type == R_MIPS_REL32:
                # 32-bit direct
                val = val + rela.addend
            elif rela_type == R_MIPS_64:
                # 64-bit direct
                val = val + rela.addend
                is_64 = True
            else:
                raise ConfigurationError(
                    f"Invalid relocation type {i} for {rela.symbol.name}: {rela_type}"
                )

        return val.to_bytes(
            8 if is_64 else 4,
            "big" if self.byteorder == platforms.Byteorder.BIG else "little",
        )


class MIPSELElfRelocator(MIPSElfRelocator):
    byteorder = platforms.Byteorder.LITTLE


class MIPS64ElfRelocator(MIPSElfRelocator):
    arch = platforms.Architecture.MIPS64


class MIPS64ELElfRelocator(MIPSElfRelocator):
    arch = platforms.Architecture.MIPS64
    byteorder = platforms.Byteorder.LITTLE
